import typing
from collections import abc

import numpy as np

from stk._internal.building_block import BuildingBlock
from stk._internal.construction_state.construction_state import (
    ConstructionState,
)
from stk._internal.optimizers.null import NullOptimizer
from stk._internal.optimizers.optimizer import Optimizer
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.topology_graphs.vertex import Vertex

from .vertices import AxleVertex, CycleVertex


class NRotaxane(TopologyGraph):
    """
    Represents [n]rotaxane topology graphs.

    Axle and cycle building blocks do not require functional groups.

    This class assumes one axle with (n-1) macrocycles threaded on it.
    The macrocycles are spaced evenly along the axle in repeating
    patterns. The threaded macrocycles can be described analogously
    to monomers in linear polymers, in terms of a repeating unit,
    except that no bonds are formed between them.

    Examples:

        *Construction*

        .. testcode:: construction

            import stk

            cycle = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='[Br]CC[Br]',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='A',
                    num_repeating_units=5,
                ),
            )
            axle = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                        stk.BuildingBlock('BrCNCBr', [stk.BromoFactory()]),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=7,
                )
            )
            rotaxane = stk.ConstructedMolecule(
                topology_graph=stk.rotaxane.NRotaxane(
                    axle=stk.BuildingBlock.init_from_molecule(axle),
                    cycles=(
                        stk.BuildingBlock.init_from_molecule(cycle),
                    ),
                    repeating_unit='A',
                    num_repeating_units=3,
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            cycle = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='[Br]CC[Br]',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='A',
                    num_repeating_units=5,
                ),
            )
            axle = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                        stk.BuildingBlock('BrCNCBr', [stk.BromoFactory()]),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=7,
                )
            )
            rotaxane = stk.ConstructedMolecule(
                topology_graph=stk.rotaxane.NRotaxane(
                    axle=stk.BuildingBlock.init_from_molecule(axle),
                    cycles=(
                        stk.BuildingBlock.init_from_molecule(cycle),
                    ),
                    repeating_unit='A',
                    num_repeating_units=3,
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        rotaxane.get_atoms(),
                        rotaxane.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in rotaxane.get_bonds()
                ),
            )

        *Suggested Optimization*

        For :class:`.NRotaxane` topologies, there is no need to use an
        optimizer, so stick with :class:`.NullOptimizer`. However, it is
        recommended that all building blocks be optimized prior to
        construction.

        *Defining the Orientation of Each Building Block*

        The `orientations` parameter allows the direction of each cycle
        along the axle to be flipped

        .. testcode:: defining-the-orientation-of-each-building-block

            import stk

            cycle1 = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='BrCNCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=5,
                    optimizer=stk.MCHammer(),
                ),
            )
            cycle2 = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='BrCCNCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=5,
                    optimizer=stk.MCHammer(),
                ),
            )
            axle = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                    ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            rotaxane1 = stk.ConstructedMolecule(
                topology_graph=stk.rotaxane.NRotaxane(
                    axle=stk.BuildingBlock.init_from_molecule(axle),
                    cycles=(
                        stk.BuildingBlock.init_from_molecule(cycle1),
                        stk.BuildingBlock.init_from_molecule(cycle2),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=3,
                    orientations=(1., 0.5),
                ),
            )

        In the above example, ``cycle1`` is guaranteed to be flipped,
        ``cycle2`` has a 50% chance of being flipped, each time it is
        placed on a node.

        Note that whether a building block will be flipped or not
        is decided during the initialization of :class:`.NRotaxane`

        .. testcode:: defining-the-orientation-of-each-building-block

            # graph1 will always construct the same [n]rotaxane.
            graph1 = stk.rotaxane.NRotaxane(
                axle=stk.BuildingBlock.init_from_molecule(axle),
                cycles=(
                    stk.BuildingBlock.init_from_molecule(cycle1),
                    stk.BuildingBlock.init_from_molecule(cycle2),
                ),
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=(0.65, 0.45),
            )
            # rotaxane2 and rotaxane3 are guaranteed to be the same as they
            # used the same topology graph.
            rotaxane2 = stk.ConstructedMolecule(graph1)
            rotaxane3 = stk.ConstructedMolecule(graph1)

            # graph2 may lead to a different [n]rotaxane, despite
            # being initialized with the same parameters.
            graph2 = stk.rotaxane.NRotaxane(
                axle=stk.BuildingBlock.init_from_molecule(axle),
                cycles=(
                    stk.BuildingBlock.init_from_molecule(cycle1),
                    stk.BuildingBlock.init_from_molecule(cycle2),
                ),
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=(0.65, 0.45)
            )

            # rotaxane4 and rotaxane5 are guaranteed to be the same because
            # they used the same topology graph. However, they may be
            # different to rotaxane2 and rotaxane3.
            rotaxane4 = stk.ConstructedMolecule(graph2)
            rotaxane5 = stk.ConstructedMolecule(graph2)

        The `random_seed` parameter can be used to get reproducible results

        .. testcode:: defining-the-orientation-of-each-building-block

            # rotaxane6 and rotaxane7 are guaranteed to be the same,
            # because graph3 and graph4 used the same random seed.

            graph3 = stk.rotaxane.NRotaxane(
                axle=stk.BuildingBlock.init_from_molecule(axle),
                cycles=(
                    stk.BuildingBlock.init_from_molecule(cycle1),
                    stk.BuildingBlock.init_from_molecule(cycle2),
                ),
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=(0.65, 0.45),
                random_seed=4,
            )
            rotaxane6 = stk.ConstructedMolecule(graph3)

            graph4 = stk.rotaxane.NRotaxane(
                axle=stk.BuildingBlock.init_from_molecule(axle),
                cycles=(
                    stk.BuildingBlock.init_from_molecule(cycle1),
                    stk.BuildingBlock.init_from_molecule(cycle2),
                ),
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=(0.65, 0.45),
                random_seed=4,
            )
            rotaxane7 = stk.ConstructedMolecule(graph4)

        *Using Numbers to Define the Repeating Unit*

        The repeating unit can also be specified through the indices of
        the building blocks

        .. testcode:: using-numbers-to-define-the-repeating-unit

            import stk

            cycle1 = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='BrCNCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=5,
                    optimizer=stk.MCHammer(),
                ),
            )
            cycle2 = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='BrCCNCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=5,
                    optimizer=stk.MCHammer(),
                ),
            )
            cycle3 = stk.ConstructedMolecule(
                topology_graph=stk.macrocycle.Macrocycle(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCNNCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='BrCCNNNCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='AB',
                    num_repeating_units=5,
                    optimizer=stk.MCHammer(),
                ),
            )
            axle = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                    ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )

            # r1 and r2 are different ways to write the same thing.
            r1 = stk.ConstructedMolecule(
                topology_graph=stk.rotaxane.NRotaxane(
                    axle=stk.BuildingBlock.init_from_molecule(axle),
                    cycles=(
                        stk.BuildingBlock.init_from_molecule(cycle1),
                        stk.BuildingBlock.init_from_molecule(cycle2),
                        stk.BuildingBlock.init_from_molecule(cycle3),
                    ),
                    repeating_unit='ACB',
                    num_repeating_units=3,
                ),
            )
            r2 = stk.ConstructedMolecule(
                topology_graph=stk.rotaxane.NRotaxane(
                    axle=stk.BuildingBlock.init_from_molecule(axle),
                    cycles=(
                        stk.BuildingBlock.init_from_molecule(cycle1),
                        stk.BuildingBlock.init_from_molecule(cycle2),
                        stk.BuildingBlock.init_from_molecule(cycle3),
                    ),
                    repeating_unit=(0, 2, 1),
                    num_repeating_units=3,
                ),
            )

    """

    def __init__(
        self,
        axle: BuildingBlock,
        cycles: abc.Iterable[BuildingBlock],
        repeating_unit: str | abc.Iterable[int],
        num_repeating_units: int,
        orientations: abc.Iterable[float] | None = None,
        random_seed: int | np.random.Generator | None = None,
        num_processes: int = 1,
        optimizer: Optimizer = NullOptimizer(),
    ) -> None:
        """
        Parameters:

            axle:
                The axle of the rotaxane.

            cycles (list[BuildingBlock]):
                The cycles threaded onto the `axle`.

            repeating_unit (str | list[int]):
                A string specifying the repeating unit of the `cycles`.
                For example, ``'AB'`` or ``'ABB'``. The first cycle in
                `cycles` is ``'A'`` and so on.

                The repeating unit can also be specified by the indices of
                `cycles`, for example ``'ABB'`` can be
                written as ``[0, 1, 1]``.

            num_repeating_units:
                The number of repeating units threaded along the axle.

            orientations (list[float] | None):
                For each character in the repeating unit, a value
                between ``0`` and ``1`` (both inclusive) must be given in
                a :class:`tuple`. It indicates the probability that each
                cycle will have its orientation along the axle
                flipped. If ``0`` then the cycle is guaranteed not to
                flip. If ``1`` it is guaranteed to flip. This allows the
                user to create head-to-head or head-to-tail chains, as well
                as chain with a preference for head-to-head or head-to-tail
                if a number between ``0`` and ``1`` is chosen. If
                ``None``, then defaults to ``0`` in every case.

                It is also possible to supply an orientation for every
                cycle vertex in the final topology graph. In this case, the
                length of `orientations` must be equal to
                ``len(repeating_unit)*num_repeating_units``.

            random_seed:
                The random seed to use when choosing random orientations.

            num_processes:
                The number of parallel processes to create during
                :meth:`construct`.

            optimizer:
                Used to optimize the structure of the constructed
                molecule.

        Raises:

            :class:`ValueError`
                If the length of `orientations` is not equal in length to
                `repeating_unit` or to the total number of vertices.

        """

        if not isinstance(repeating_unit, str):
            repeating_unit = tuple(repeating_unit)

        if orientations is None:
            orientations = tuple(0.0 for _ in range(len(repeating_unit)))
        else:
            orientations = tuple(orientations)

        if len(orientations) == len(repeating_unit):
            orientations = orientations * num_repeating_units

        chain_length = len(repeating_unit) * num_repeating_units
        if len(orientations) != chain_length:
            raise ValueError(
                "The length of orientations must match either "
                "the length of repeating_unit or the "
                "total number of vertices."
            )

        if random_seed is None or isinstance(random_seed, int):
            random_seed = np.random.default_rng(random_seed)

        self._repeating_unit = self._normalize_repeating_unit(
            repeating_unit=repeating_unit,
        )
        self._num_repeating_units = num_repeating_units

        vertices: list[AxleVertex | CycleVertex] = [AxleVertex(0, (0, 0, 0))]
        distance = 1 / (chain_length + 1)
        choices = [True, False]
        for vertex_id, flip_chance in enumerate(orientations, 1):
            vertices.append(
                CycleVertex(
                    id=vertex_id,
                    position=[vertex_id * distance - 0.5, 0, 0],
                    flip=random_seed.choice(
                        choices,
                        p=[flip_chance, 1 - flip_chance],
                    ),
                )
            )

        # Save the chosen orientations for __repr__.
        self._orientations = tuple(
            int(vertex.get_flip())  # type: ignore[union-attr]
            for vertex in vertices[1:]
        )

        super().__init__(
            building_block_vertices=self._get_building_block_vertices(
                axle=axle,
                cycles=tuple(cycles),
                vertices=tuple(vertices),
            ),
            edges=(),
            reaction_factory=GenericReactionFactory(),
            construction_stages=(),
            num_processes=num_processes,
            optimizer=optimizer,
            edge_groups=None,
        )

    def clone(self) -> typing.Self:
        clone = super().clone()
        clone._repeating_unit = self._repeating_unit
        clone._num_repeating_units = self._num_repeating_units
        clone._orientations = self._orientations
        return clone

    @staticmethod
    def _normalize_repeating_unit(
        repeating_unit: str | tuple[int, ...],
    ) -> tuple[int, ...]:
        if isinstance(repeating_unit, tuple):
            return repeating_unit
        base = ord("A")
        return tuple(ord(letter) - base for letter in repeating_unit)

    def _get_building_block_vertices(
        self,
        axle: BuildingBlock,
        cycles: abc.Sequence[BuildingBlock],
        vertices: abc.Sequence[Vertex],
    ) -> dict[BuildingBlock, abc.Sequence[Vertex]]:
        threads = self._repeating_unit * self._num_repeating_units
        building_block_vertices: dict[BuildingBlock, list[Vertex]] = {}

        building_block_vertices[axle] = [vertices[0]]
        for cycle_index, vertex in zip(threads, vertices[1:]):
            bb = cycles[cycle_index]
            building_block_vertices[bb] = building_block_vertices.get(bb, [])
            building_block_vertices[bb].append(vertex)

        return typing.cast(
            dict[BuildingBlock, abc.Sequence[Vertex]],
            building_block_vertices,
        )

    def _run_reactions(self, state: ConstructionState) -> ConstructionState:
        return state

    @staticmethod
    def _get_scale(
        building_block_vertices: dict[BuildingBlock, abc.Sequence[Vertex]],
        scale_multiplier: float,
    ) -> float:
        axle = next(iter(building_block_vertices))
        return 0.8 * axle.get_maximum_diameter()

    def __repr__(self):
        return (
            f"rotaxane.NRotaxane("
            f"{self._repeating_unit!r}, "
            f"{self._num_repeating_units}, "
            f"{self._orientations!r}"
            f")"
        )



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def secondary_amino(get_atom_ids):
    a, b, c, d = get_atom_ids(4)
    return _secondary_amino(stk.N(a), stk.H(b), stk.C(c), stk.C(d))


def _secondary_amino(nitrogen, hydrogen, atom1, atom2):
    bonders = (nitrogen,)
    deleters = (hydrogen,)
    return GenericCaseData(
        functional_group=stk.SecondaryAmino(
            nitrogen=nitrogen,
            hydrogen=hydrogen,
            atom1=atom1,
            atom2=atom2,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(nitrogen, hydrogen, atom1, atom2),
        bonders=bonders,
        deleters=deleters,
    )



import stk


def get_pd_atom() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles="[Pd+2]",
        functional_groups=(
            stk.SingleAtom(stk.Pd(0, charge=2)) for i in range(4)
        ),
        position_matrix=([0, 0, 0],),
    )


def get_fe_atom() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles="[Fe+2]",
        functional_groups=(
            stk.SingleAtom(stk.Fe(0, charge=2)) for i in range(6)
        ),
        position_matrix=([0, 0, 0],),
    )


def get_linker() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles=(
            "[H]C1=NC([H])=C([H])C(C2=C([H])C([H])=C([H])C(C3=C([H])"
            "C([H])=NC([H])=C3[H])=C2[H])=C1[H]"
        ),
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts="[#6]~[#7X2]~[#6]",
                bonders=(1,),
                deleters=(),
            ),
        ],
    )


def get_other_linker() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles=("[H]C1=NC([H])=C([H])C(C2=C([H])C([H])=NC([H])=C2[H])=C1[H]"),
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts="[#6]~[#7X2]~[#6]",
                bonders=(1,),
                deleters=(),
            ),
        ],
    )


def get_palladium_bi_1() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles="[H]N([H])C([H])([H])C([H])([H])N([H])[H]",
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts="[#7]~[#6]",
                bonders=(0,),
                deleters=(),
            ),
        ],
    )


def get_palladium_cispbi_sqpl() -> stk.BuildingBlock:
    molecule = stk.ConstructedMolecule(
        topology_graph=stk.metal_complex.CisProtectedSquarePlanar(
            metals={get_pd_atom(): 0},
            ligands={get_palladium_bi_1(): 0},
            reaction_factory=stk.DativeReactionFactory(
                reaction_factory=stk.GenericReactionFactory(
                    bond_orders={
                        frozenset(
                            {
                                stk.GenericFunctionalGroup,
                                stk.SingleAtom,
                            }
                        ): 9,
                    },
                ),
            ),
        ),
    )

    return stk.BuildingBlock.init_from_molecule(
        molecule=molecule,
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts="[Pd]~[#7]",
                bonders=(0,),
                deleters=(),
                placers=(0, 1),
            ),
        ],
    )


def get_ditopic_linker() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles=(
            "[H]C1=C([H])C(C2=C([H])C([H])=C(Br)C([H])=C2[H])=C([H])"
            "C([H])=C1Br"
        ),
        functional_groups=[stk.BromoFactory()],
    )


def get_tritopic_linker() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles=(
            "[H]C1=C([H])C(N(C2=C([H])C([H])=C(Br)C([H])=C2[H])"
            "C2=C([H])C("
            "[H])=C(Br)C([H])=C2[H])=C([H])C([H])=C1Br"
        ),
        functional_groups=[stk.BromoFactory()],
    )


def get_tetratopic_linker() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles=(
            "[H]C1=C([H])C(C(C2=C([H])C([H])=C(Br)C([H])=C2[H])C"
            "(C2=C([H])C([H])=C(Br)C([H])=C2[H])C2=C([H])C([H])=C"
            "(Br)C([H])=C2[H])=C([H])C([H])=C1Br"
        ),
        functional_groups=[stk.BromoFactory()],
    )


def get_complex_ligand() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles="[H]C1=NC(C([H])=NBr)=C([H])C([H])=C1[H]",
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts="[#6]~[#7X2]~[#35]",
                bonders=(1,),
                deleters=(),
            ),
            stk.SmartsFunctionalGroupFactory(
                smarts="[#6]~[#7X2]~[#6]",
                bonders=(1,),
                deleters=(),
            ),
        ],
    )


def get_iron_complex() -> stk.BuildingBlock:
    iron_complex = stk.ConstructedMolecule(
        topology_graph=stk.metal_complex.OctahedralDelta(
            metals={get_fe_atom(): 0},
            ligands={get_complex_ligand(): (0, 1, 2)},
            reaction_factory=stk.DativeReactionFactory(
                reaction_factory=stk.GenericReactionFactory(
                    bond_orders={
                        frozenset(
                            {
                                stk.GenericFunctionalGroup,
                                stk.SingleAtom,
                            }
                        ): 9,
                    }
                )
            ),
        )
    )
    return stk.BuildingBlock.init_from_molecule(
        molecule=iron_complex, functional_groups=[stk.BromoFactory()]
    )


def get_iron_bi_1() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles="BrN=Cc1ccccn1",
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts="[#6]~[#7X2]~[#35]",
                bonders=(1,),
                deleters=(),
            ),
            stk.SmartsFunctionalGroupFactory(
                smarts="[#6]~[#7X2]~[#6]",
                bonders=(1,),
                deleters=(),
            ),
        ],
    )


def get_mo_1() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles="c1cc2c(cn1)CCCCC2",
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts="[#6]~[#7X2]~[#6]",
                bonders=(1,),
                deleters=(),
            ),
        ],
    )



"""
Diol
====

"""

from .generic_functional_group import GenericFunctionalGroup


class Diol(GenericFunctionalGroup):
    """
    Represents a diol functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[hydrogen1][oxygen1][atom1][atom2][oxygen2][hydrogen2]``.

    """

    def __init__(
        self,
        atom1,
        oxygen1,
        hydrogen1,
        atom2,
        oxygen2,
        hydrogen2,
        bonders,
        deleters,
        placers=None,
    ):
        """
        Initialize a :class:`.Diol` instance.

        Parameters
        ----------
        atom1 : :class:`.Atom`
            The ``[atom1]`` atom.

        oxygen1 : :class:`.O`
            The ``[oxygen1]`` atom.

        hydrogen1 : :class:`.H`
            The ``[hydrogen1]`` atom.

        atom2 : :class:`.Atom`
            The ``[atom2]`` atom.

        oxygen2 : :class:`.O`
            The ``[oxygen2]`` atom.

        hydrogen2 : :class:`.H`
            The ``[hydrogen2]`` atom.

        bonders : :class:`tuple` of :class:`.Atom`
            The bonder atoms.

        deleters : :class:`tuple` of :class:`.Atom`
            The deleter atoms.

        placers : :class:`tuple` of :class:`.Atom`, optional
            The placer atoms. If ``None`` the `bonders` will be used.

        """

        self._atom1 = atom1
        self._oxygen1 = oxygen1
        self._hydrogen1 = hydrogen1
        self._atom2 = atom2
        self._oxygen2 = oxygen2
        self._hydrogen2 = hydrogen2
        atoms = (atom1, oxygen1, hydrogen1, atom2, oxygen2, hydrogen2)
        super().__init__(
            atoms=atoms,
            bonders=bonders,
            deleters=deleters,
            placers=bonders if placers is None else placers,
        )

    def get_atom1(self):
        """
        Get the ``[atom1]`` atom.

        Returns
        -------
        :class:`.Atom`
            The ``[atom1]`` atom.

        """

        return self._atom1

    def get_oxygen1(self):
        """
        Get the ``[oxygen1]`` atom.

        Returns
        -------
        :class:`.O`
            The ``[oxygen1]`` atom.

        """

        return self._oxygen1

    def get_hydrogen1(self):
        """
        Get the ``[hydrogen1]`` atom.

        Returns
        -------
        :class:`.H`
            The ``[hydrogen1]`` atom.

        """

        return self._hydrogen1

    def get_atom2(self):
        """
        Get the ``[atom2]`` atom.

        Returns
        -------
        :class:`.Atom`
            The ``[atom2]`` atom.

        """

        return self._atom2

    def get_oxygen2(self):
        """
        Get the ``[oxygen2]`` atom.

        Returns
        -------
        :class:`.O`
            The ``[oxygen2]`` atom.

        """

        return self._oxygen2

    def get_hydrogen2(self):
        """
        Get the ``[hydrogen2]`` atom.

        Returns
        -------
        :class:`.H`
            The ``[hydrogen2]`` atom.

        """

        return self._hydrogen2

    def clone(self):
        clone = super().clone()
        clone._atom1 = self._atom1
        clone._oxygen1 = self._oxygen1
        clone._hydrogen1 = self._hydrogen1
        clone._atom2 = self._atom2
        clone._oxygen2 = self._oxygen2
        clone._hydrogen2 = self._hydrogen2
        return clone

    def with_atoms(self, atom_map):
        clone = super().with_atoms(atom_map)
        clone._atom1 = atom_map.get(
            self._atom1.get_id(),
            self._atom1,
        )
        clone._oxygen1 = atom_map.get(
            self._oxygen1.get_id(),
            self._oxygen1,
        )
        clone._hydrogen1 = atom_map.get(
            self._hydrogen1.get_id(),
            self._hydrogen1,
        )
        clone._atom2 = atom_map.get(
            self._atom2.get_id(),
            self._atom2,
        )
        clone._oxygen2 = atom_map.get(
            self._oxygen2.get_id(),
            self._oxygen2,
        )
        clone._hydrogen2 = atom_map.get(
            self._hydrogen2.get_id(),
            self._hydrogen2,
        )
        return clone

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{self._atom1}, {self._oxygen1}, {self._hydrogen1}, "
            f"{self._atom2}, {self._oxygen2}, {self._hydrogen2}, "
            f"bonders={self._bonders}, deleters={self._deleters})"
        )



import logging

import numpy as np

from stk._internal.building_block import BuildingBlock
from stk._internal.topology_graphs.vertex import Vertex
from stk._internal.utilities.utilities import get_acute_vector

from ..edge import Edge
from ..utilities import _EdgeSorter, _FunctionalGroupSorter

logger = logging.getLogger(__name__)


class CoreVertex(Vertex):
    """Represents a vertex in the core of an ncore topology graph."""

    def place_building_block(
        self,
        building_block: BuildingBlock,
        edges: tuple[Edge, ...],
    ) -> np.ndarray:
        # Sets building block in XY plane, then aligns 0th functional
        # group with the 0th edge.
        assert building_block.get_num_functional_groups() > 2, (
            f"{building_block} needs to have more than 2 functional "
            "groups but has "
            f"{building_block.get_num_functional_groups()}."
        )
        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        normal = building_block.get_plane_normal(
            atom_ids=building_block.get_placer_ids(),
        )
        normal = get_acute_vector(
            reference=core_centroid - self._position,
            vector=normal,
        )
        building_block = building_block.with_rotation_between_vectors(
            start=normal,
            target=np.array([0.0, 0.0, 1.0]),
            origin=self._position,
        )
        (fg,) = building_block.get_functional_groups(0)
        fg_centroid = building_block.get_centroid(fg.get_placer_ids())
        edge_position = edges[0].get_position()
        return building_block.with_rotation_to_minimize_angle(
            start=fg_centroid - self._position,
            target=edge_position - self._position,
            axis=np.array([0, 0, 1], dtype=np.float64),
            origin=self._position,
        ).get_position_matrix()

    def map_functional_groups_to_edges(
        self,
        building_block: BuildingBlock,
        edges: tuple[Edge, ...],
    ) -> dict[int, int]:
        # The idea is to order the functional groups in building_block
        # by their angle with the vector running from the placer
        # centroid to fg0, going in the clockwise direction.
        # The edges are also ordered by their angle with the vector
        # running from the edge centroid to the aligner_edge,
        # going in the clockwise direction.
        #
        # Once the fgs and edges are ordered, zip and assign them.

        fg_sorter = _FunctionalGroupSorter(building_block)
        edge_sorter = _EdgeSorter(
            edges=edges,
            aligner_edge=edges[0],
            axis=fg_sorter.get_axis(),
        )
        return {
            fg_id: edge.get_id()
            for fg_id, edge in zip(
                fg_sorter.get_items(),
                edge_sorter.get_items(),
            )
        }


class SubstituentVertex(Vertex):
    """Represents a vertex to be bound to core."""

    def place_building_block(
        self,
        building_block: BuildingBlock,
        edges: tuple[Edge, ...],
    ) -> np.ndarray:
        assert building_block.get_num_functional_groups() == 1, (
            f"{building_block} needs to have 1 functional "
            "groups but has "
            f"{building_block.get_num_functional_groups()}."
        )

        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_core_atom_ids(),
        )

        (fg,) = building_block.get_functional_groups()
        fg_centroid = building_block.get_centroid(
            atom_ids=fg.get_placer_ids(),
        )
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        edge_centroid = sum(edge.get_position() for edge in edges) / len(edges)
        return building_block.with_rotation_between_vectors(
            start=(fg_centroid - core_centroid),
            target=edge_centroid - self._position,
            origin=self._position,
        ).get_position_matrix()

    def map_functional_groups_to_edges(
        self,
        building_block: BuildingBlock,
        edges: tuple[Edge, ...],
    ) -> dict[int, int]:
        return {0: edges[0].get_id()}


class SingleVertex(Vertex):
    """Represents a vertex for internal reaction placed at the origin."""

    def place_building_block(
        self,
        building_block: BuildingBlock,
        edges: tuple[Edge, ...],
    ) -> np.ndarray:
        return building_block.with_centroid(
            np.array([0.0, 0.0, 0.0])
        ).get_position_matrix()

    def map_functional_groups_to_edges(
        self,
        building_block: BuildingBlock,
        edges: tuple[Edge, ...],
    ) -> dict[int, int]:
        # The idea is to order the functional groups in building_block
        # by their angle with the vector running from the placer
        # centroid to fg0, going in the clockwise direction.
        # The edges are also ordered by their angle with the vector
        # running from the edge centroid to the aligner_edge,
        # going in the clockwise direction.
        #
        # Once the fgs and edges are ordered, zip and assign them.

        fg_sorter = _FunctionalGroupSorter(building_block)
        edge_sorter = _EdgeSorter(
            edges=edges,
            aligner_edge=edges[0],
            axis=fg_sorter.get_axis(),
        )
        return {
            fg_id: edge.get_id()
            for fg_id, edge in zip(
                fg_sorter.get_items(),
                edge_sorter.get_items(),
            )
        }



import os
import pathlib

import pytest


@pytest.fixture(
    params=[
        "molecule.mol",
        "molecule.xyz",
        pathlib.Path("molecule-path.mol"),
        pathlib.Path("molecule-path.xyz"),
    ],
)
def path(request, tmpdir):
    return os.path.join(tmpdir, request.param)



import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(stk.C(0),),
)
def atom(request):
    """
    A :class:`.Atom` instance.

    """

    return request.param


@pytest.fixture(
    params=(
        stk.C(0),
        None,
    ),
)
def building_block_atom(request):
    """
    A valid building block atom for :class:`.AtomInfo`.

    """

    return request.param


@pytest.fixture(
    params=(
        stk.BuildingBlock("NCCN"),
        None,
    ),
)
def building_block(request):
    """
    A valid building block for :class:`.AtomInfo`.

    """

    return request.param


@pytest.fixture(
    params=(
        1,
        None,
    ),
)
def building_block_id(request):
    """
    A building block id.

    """

    return request.param


@pytest.fixture
def case_data(
    atom,
    building_block_atom,
    building_block,
    building_block_id,
):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(
        atom=atom,
        building_block_atom=building_block_atom,
        building_block=building_block,
        building_block_id=building_block_id,
    )



import typing
from collections import abc
from dataclasses import dataclass

import numpy as np

from stk._internal.building_block import BuildingBlock
from stk._internal.optimizers.null import NullOptimizer
from stk._internal.optimizers.optimizer import Optimizer
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.reaction_factories.reaction_factory import (
    ReactionFactory,
)
from stk._internal.topology_graphs.edge import Edge
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.topology_graphs.vertex import Vertex

from .vertices import CoreVertex, SubstituentVertex


class NCore(TopologyGraph):
    """
    Represents a core reacting with N arms topology graph.

    The core building block, with N functional groups, defines the
    behaviour of this topology graph. The arm building blocks, which
    must have only one functional group, are oriented in a circle around
    the core based on N and the repeating units. The core is oriented
    into the XY plane.

    Examples:

        *Construction*

        The number of functional groups in the core building block
        define the topology graph

        .. testcode:: construction

            import stk

            core1 = stk.BuildingBlock(
                smiles="C(Br)1C(Br)C1Br",
                functional_groups=stk.BromoFactory(),
            )
            arm = stk.BuildingBlock(
                smiles="BrC",
                functional_groups=stk.BromoFactory(),
            )

            ncore = stk.ConstructedMolecule(stk.small.NCore(
                core_building_block=core1,
                arm_building_blocks=arm,
                repeating_unit="A",  # 'AAA' would work too.
            ))

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            arm = stk.BuildingBlock(
                smiles="BrC",
                functional_groups=stk.BromoFactory(),
            )
            core1 = stk.BuildingBlock(
                smiles="C(Br)1C(Br)C1Br",
                functional_groups=stk.BromoFactory(),
            )

            ncore = stk.ConstructedMolecule(stk.small.NCore(
                core_building_block=core1,
                arm_building_blocks=arm,
                repeating_unit="A",
            ))

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        ncore.get_atoms(),
                        ncore.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in ncore.get_bonds()
                ),
            )

        .. testcode:: construction

            import stk

            core2 = stk.BuildingBlock(
                smiles="C(Br)1C(Br)C(Br)C(Br)C(Br)C(Br)C1Br",
                functional_groups=stk.BromoFactory(),
            )
            arm = stk.BuildingBlock(
                smiles="BrC",
                functional_groups=stk.BromoFactory(),
            )

            ncore = stk.ConstructedMolecule(stk.small.NCore(
                core_building_block=core2,
                arm_building_blocks=arm,
                repeating_unit="A",  # 'AAAAAAA' would work too.
            ))

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            arm = stk.BuildingBlock(
                smiles="BrC",
                functional_groups=stk.BromoFactory(),
            )
            core2 = stk.BuildingBlock(
                smiles="C(Br)1C(Br)C(Br)C(Br)C(Br)C(Br)C1Br",
                functional_groups=stk.BromoFactory(),
            )

            ncore = stk.ConstructedMolecule(stk.small.NCore(
                core_building_block=core2,
                arm_building_blocks=arm,
                repeating_unit="A",
            ))

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        ncore.get_atoms(),
                        ncore.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in ncore.get_bonds()
                ),
            )

        *Suggested Optimization*

        For :class:`.NCore` topologies, it is recommend to use the
        :class:`.MCHammer` optimizer.

        .. testcode:: suggested-optimization

            import stk

            arm = stk.BuildingBlock(
                smiles="BrC",
                functional_groups=stk.BromoFactory(),
            )
            core2 = stk.BuildingBlock(
                smiles="C(Br)1C(Br)C(Br)C(Br)C(Br)C(Br)C1Br",
                functional_groups=stk.BromoFactory(),
            )

            ncore = stk.ConstructedMolecule(
                stk.small.NCore(
                    core_building_block=core2,
                    arm_building_blocks=arm,
                    repeating_unit="A",
                    optimizer=stk.MCHammer(),
                )
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            arm = stk.BuildingBlock(
                smiles="BrC",
                functional_groups=stk.BromoFactory(),
            )
            core2 = stk.BuildingBlock(
                smiles="C(Br)1C(Br)C(Br)C(Br)C(Br)C(Br)C1Br",
                functional_groups=stk.BromoFactory(),
            )

            ncore = stk.ConstructedMolecule(
                stk.small.NCore(
                    core_building_block=core2,
                    arm_building_blocks=arm,
                    repeating_unit="A",
                    optimizer=stk.MCHammer(),
                )
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        ncore.get_atoms(),
                        ncore.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in ncore.get_bonds()
                ),
            )

        *Construction with Distinct Arms*

        Any combination of arm building blocks can be provided.

        .. testcode:: construction-with-distinct-arms1

            import stk

            core = stk.BuildingBlock(
                smiles="C(Br)1C(Br)C(Br)C(Br)C(Br)C(Br)C1Br",
                functional_groups=stk.BromoFactory(),
            )
            arm1 = stk.BuildingBlock(
                smiles="BrC",
                functional_groups=stk.BromoFactory(),
            )
            arm2 = stk.BuildingBlock(
                smiles="BrCN",
                functional_groups=stk.BromoFactory(),
            )

            ncore = stk.ConstructedMolecule(
                stk.small.NCore(
                    core_building_block=core,
                    arm_building_blocks=[arm1, arm2],
                    repeating_unit="ABABABA",
                )
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            core = stk.BuildingBlock(
                smiles="C(Br)1C(Br)C(Br)C(Br)C(Br)C1Br",
                functional_groups=stk.BromoFactory(),
            )
            arm1 = stk.BuildingBlock(
                smiles="BrC",
                functional_groups=stk.BromoFactory(),
            )
            arm2 = stk.BuildingBlock(
                smiles="BrCN",
                functional_groups=stk.BromoFactory(),
            )

            ncore = stk.ConstructedMolecule(
                stk.small.NCore(
                    core_building_block=core,
                    arm_building_blocks=[arm1, arm2],
                    repeating_unit="ABABAB",
                )
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        ncore.get_atoms(),
                        ncore.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in ncore.get_bonds()
                ),
            )

        Just be cautious of the repeating unit term, which must match
        the available functional groups in the core. Units that are a
        subset of available functional groups will be repeated.

        .. testcode:: construction-with-distinct-arms2

            import stk

            core = stk.BuildingBlock(
                smiles="C(Br)1C(Br)C(Br)C(Br)C(Br)C1Br",
                functional_groups=stk.BromoFactory(),
            )
            arm1 = stk.BuildingBlock(
                smiles="BrC",
                functional_groups=stk.BromoFactory(),
            )
            arm2 = stk.BuildingBlock(
                smiles="BrCN",
                functional_groups=stk.BromoFactory(),
            )

            ncore = stk.ConstructedMolecule(
                stk.small.NCore(
                    core_building_block=core,
                    arm_building_blocks=[arm1, arm2],
                    repeating_unit="ABA",  # Same as 'ABAABA'.
                )
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            core = stk.BuildingBlock(
                smiles="C(Br)1C(Br)C(Br)C(Br)C(Br)C1Br",
                functional_groups=stk.BromoFactory(),
            )
            arm1 = stk.BuildingBlock(
                smiles="BrC",
                functional_groups=stk.BromoFactory(),
            )
            arm2 = stk.BuildingBlock(
                smiles="BrCN",
                functional_groups=stk.BromoFactory(),
            )

            ncore = stk.ConstructedMolecule(
                stk.small.NCore(
                    core_building_block=core,
                    arm_building_blocks=[arm1, arm2],
                    repeating_unit="ABA",
                )
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        ncore.get_atoms(),
                        ncore.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in ncore.get_bonds()
                ),
            )

    """

    def __init__(
        self,
        core_building_block: BuildingBlock,
        arm_building_blocks: BuildingBlock | abc.Iterable[BuildingBlock],
        repeating_unit: str | abc.Iterable[int],
        reaction_factory: ReactionFactory = GenericReactionFactory(),
        num_processes: int = 1,
        optimizer: Optimizer = NullOptimizer(),
        scale_multiplier: float = 1.0,
    ) -> None:
        """
        Parameters:

            core_building_block:
                The central building block.

            arm_building_blocks (BuildingBlock | list[BuildingBlock]):
                The building blocks to react with the core.

            repeating_unit (str | list[int]):
                A string specifying the repeating unit of the arms.
                For example, ``'AB'`` or ``'ABB'``. The first building
                block passed to `arm_building_blocks` is ``'A'`` and
                so on. The repeating unit must fit evenly onto the
                `core_building_block`, defined by the number of
                functional groups the core has.

                The repeating unit can also be specified by the
                indices of `building_blocks`, for example ``'ABB'``
                can be written as ``[0, 1, 1]``.

            reaction_factory:
                The factory to use for creating reactions between
                functional groups of building blocks.

            num_processes:
                The number of parallel processes to create during
                :meth:`construct`.

            optimizer:
                Used to optimize the structure of the constructed
                molecule.

            scale_multiplier:
                Scales the positions of the vertices.

        Raises:

            :class:`ValueError`
                If the length of `orientations` is not equal in length
                to `repeating_unit` or to the total number of vertices.

        """

        if isinstance(arm_building_blocks, BuildingBlock):
            arm_building_blocks = (arm_building_blocks,)

        self._repr = (
            f"NCore({core_building_block!r}, {arm_building_blocks!r}, "
            f"{repeating_unit!r})"
        )

        if not isinstance(repeating_unit, str):
            repeating_unit = tuple(repeating_unit)

        self._num_arms = core_building_block.get_num_functional_groups()
        if self._num_arms % len(repeating_unit) != 0:
            raise ValueError(
                f"The repeating unit {repeating_unit} does not fit "
                "evenly onto the core building block with "
                f"{self._num_arms} functional groups."
            )

        # Keep these for __repr__.
        self._repeating_unit = self._normalize_repeating_unit(
            repeating_unit=repeating_unit
        )
        self._num_repeating_units = int(self._num_arms / len(repeating_unit))

        vertices_and_edges = self._get_vertices_and_edges(self._num_arms)
        vertices = vertices_and_edges.vertices
        edges = vertices_and_edges.edges

        super().__init__(
            building_block_vertices=self._get_building_block_vertices(
                core_building_block=core_building_block,
                arm_building_blocks=tuple(arm_building_blocks),
                vertices=vertices,
            ),
            edges=edges,
            reaction_factory=reaction_factory,
            construction_stages=(),
            optimizer=optimizer,
            num_processes=num_processes,
            scale_multiplier=scale_multiplier,
        )

    @staticmethod
    def _get_vertices_and_edges(num_arms: int) -> "_VerticesAndEdges":
        """
        Get the vertices and edges of the topology graph.

        Parameters:

            num_arms:
                The number of arms to add to the core.

        Returns:

            The vertices and edges of the topology graph.

        """

        if num_arms == 1:
            arm_positions = [np.array((1, 0, 0))]
        elif num_arms == 2:
            arm_positions = [np.array((1, 0, 0)), np.array((-1, 0, 0))]
        else:
            thetas = np.linspace(0, 2 * np.pi, num_arms + 1)[:-1]
            x_points = np.cos(thetas)
            y_points = np.sin(thetas)
            arm_positions = [
                np.array((x, y, 0)) for x, y in zip(x_points, y_points)
            ]

        vertices: list[Vertex] = [CoreVertex(0, (0, 0, 0))]
        edges: list[Edge] = []
        for i, pos in enumerate(arm_positions):
            vertices.append(SubstituentVertex(i + 1, pos))
            edges.append(Edge(len(edges), vertices[0], vertices[-1]))

        return _VerticesAndEdges(
            vertices=tuple(vertices),
            edges=tuple(edges),
        )

    def clone(self) -> typing.Self:
        clone = self._clone()
        clone._repr = self._repr
        clone._repeating_unit = self._repeating_unit
        clone._num_repeating_units = self._num_repeating_units
        return clone

    @staticmethod
    def _normalize_repeating_unit(
        repeating_unit: str | tuple[int, ...],
    ) -> tuple[int, ...]:
        if isinstance(repeating_unit, tuple):
            return repeating_unit

        base = ord("A")
        return tuple(ord(letter) - base for letter in repeating_unit)

    def _get_building_block_vertices(
        self,
        core_building_block: BuildingBlock,
        arm_building_blocks: tuple[BuildingBlock, ...],
        vertices: tuple[Vertex, ...],
    ) -> dict[BuildingBlock, abc.Sequence[Vertex]]:
        building_block_vertices: dict[BuildingBlock, list[Vertex]] = {}
        building_block_vertices[core_building_block] = [vertices[0]]

        bb_order = self._repeating_unit * self._num_repeating_units
        for bb_index, vertex in zip(bb_order, vertices[1:]):
            bb = arm_building_blocks[bb_index]
            building_block_vertices[bb] = building_block_vertices.get(bb, [])
            building_block_vertices[bb].append(vertex)

        return {
            i: tuple(building_block_vertices[i])
            for i in building_block_vertices
        }

    @staticmethod
    def _get_scale(
        building_block_vertices: dict[BuildingBlock, abc.Sequence[Vertex]],
        scale_multiplier: float,
    ) -> float:
        return scale_multiplier * max(
            bb.get_maximum_diameter() for bb in building_block_vertices
        )

    def with_building_blocks(
        self,
        building_block_map: dict[BuildingBlock, BuildingBlock],
    ) -> typing.Self:
        return self.clone()._with_building_blocks(building_block_map)

    def __repr__(self) -> str:
        return self._repr


@dataclass(frozen=True)
class _VerticesAndEdges:
    vertices: tuple[Vertex, ...]
    edges: tuple[Edge, ...]



"""
Cage Construction State
=======================

"""

from stk._internal.construction_state import ConstructionState


class _CageConstructionState(ConstructionState):
    """
    The construction state of a :class:`.Cage`.

    """

    def __init__(
        self,
        building_block_vertices,
        edges,
        num_placement_stages,
        vertex_degrees,
    ):
        """
        Initialize a :class:`._CageConstructionState` instance.

        Parameters
        ----------
        building_block_vertices : :class:`dict`
            Maps each :class:`.BuildingBlock` to be placed, to a
            :class:`tuple` of :class:`.Vertex` instances, on which
            it should be placed.

        edges : :class:`tuple` of :class:`.Edge`
            The edges of the topology graph.

        num_placement_stages : :class:`int`
            The number of placement stages.

        vertex_degrees : :class:`dict`
            Maps the id of a vertex to the number of edges it is
            connected to.

        """

        super().__init__(
            building_block_vertices=building_block_vertices,
            edges=edges,
            lattice_constants=(),
        )
        self._num_placement_stages = num_placement_stages
        self._num_placement_stages_done = 0
        self._vertex_degrees = dict(vertex_degrees)
        self._neighbor_positions = {}

    def _with_placement_results(
        self,
        vertices,
        edges,
        building_blocks,
        results,
    ):
        # Need to iterate multiple times through results.
        results = tuple(results)
        super()._with_placement_results(
            vertices=vertices,
            edges=edges,
            building_blocks=building_blocks,
            results=results,
        )

        # No need to update vertex positions after the last stage.
        self._num_placement_stages_done += 1
        if self._num_placement_stages_done >= self._num_placement_stages:
            return self

        self._update_neighbor_positions(
            vertices=vertices,
            edges=edges,
            building_blocks=building_blocks,
            results=results,
        )
        return self

    def _update_neighbor_positions(
        self,
        vertices,
        edges,
        building_blocks,
        results,
    ):
        # Use literal docstring here to prevent linter errors stemming
        # from the use of "\" in the docstring.
        r"""
        Put vertices in the middle of placer centroids.

        Normally, linear vertices are in the middle of the
        non-linear vertices they are neighbors to. However,
        they should really be placed in the middle of the functional
        groups sitting on the vertices. This ensures that the bonds
        made are straight.

        For example, take ``N`` to be a non-linear vertex, ``L``
        to be a linear vertex and ``X`` to be a functional group
        placed on a vertex. If, vertex positions are not updated,
        a linear vertex will be in the middle of the two non-linear
        vertices:

            X               X
            |               |
            N    X--L--X    N

        Therefore, the bonds made will be bent:

            X           X
            | \       / |
            N  X--L--X  N

        However, this method updates the position of ``L`` to
        be in the middle of the functional groups:

            X    X--L--X    X
            |               |
            N               N

        Now when bonds are made, they will be straight:

            X----X--L--X----X
            |               |
            N               N

        Note that all vertices, which can have their positions updated,
        will have them updated, even the non-linear ones. A vertex
        can have its position updated, if the positions of all of the
        neighboring functional groups (the ``X`` in the diagrams
        above), are known.

        Parameters
        ----------
        vertices : :class:`iterable` of :class:`vertex`
            The vertices which were just used to place a set of
            `building_blocks`.

        edges : :class:`iterable`
            For each vertex in `vertices`, a :class:`tuple` of of the
            :class:`.Edge` instances it is connected to.

        building_blocks : :class:`tuple` of :class:`.BuildingBlock`
            The building blocks which are just placed.

        results : :class:`iterable` of :class:`._PlacementResult`
            For each vertex in `vertices`, the result of the placement.

        Returns
        -------
        None : :class:`NoneType`

        """

        for vertex, vertex_edges, building_block, result in zip(
            vertices,
            edges,
            building_blocks,
            results,
        ):
            building_block = building_block.with_position_matrix(
                position_matrix=result.position_matrix,
            )
            edge_functional_groups = dict(
                zip(
                    result.functional_group_edges.values(),
                    result.functional_group_edges.keys(),
                )
            )
            for neighbor_id, edge_id in self._get_neighbors(
                vertex=vertex,
                vertex_edges=vertex_edges,
            ):
                fg_id = edge_functional_groups[edge_id]
                (functional_group,) = building_block.get_functional_groups(
                    fg_id
                )
                self._neighbor_positions[neighbor_id] = (
                    self._neighbor_positions.get(neighbor_id, [])
                )
                self._neighbor_positions[neighbor_id].append(
                    building_block.get_centroid(
                        atom_ids=functional_group.get_placer_ids(),
                    )
                )

        self._graph_state = self._graph_state.with_vertices(
            vertices=self._get_new_vertices(),
        )

    def _get_neighbors(self, vertex, vertex_edges):
        """
        Yield the neighbor vertices of `vertex`.

        Parameters
        ----------
        vertex : :class:`.Vertex`
            The vertex whose neighbors are desired.

        vertex_edges : :class:`tuple` of :class:`.Edge`
            The edges connected to `vertex`.

        Yields
        ------
        :class:`tuple`
            The first element is the id of a neighbor vertex and
            the second element is the id of the edge through which
            it is connected.

        """

        for edge in vertex_edges:
            neighbor_id = (
                edge.get_vertex1_id()
                if vertex.get_id() != edge.get_vertex1_id()
                else edge.get_vertex2_id()
            )
            (neighbor,) = self._graph_state.get_vertices(neighbor_id)
            if neighbor.use_neighbor_placement():
                yield neighbor_id, edge.get_id()

    def _get_new_vertices(self):
        """
        Yield the vertices once new positions have been added.

        Yields
        ------
        :class:`.Vertex`
            A vertex of the topology graph.

        """

        for vertex_id, vertex in enumerate(self._graph_state.get_vertices()):
            if (
                len(self._neighbor_positions.get(vertex_id, []))
                == self._vertex_degrees[vertex_id]
            ):
                yield vertex.with_position(
                    position=(
                        sum(self._neighbor_positions[vertex_id])
                        / len(self._neighbor_positions[vertex_id])
                    ),
                )
            else:
                yield vertex

    def clone(self):
        clone = super().clone()
        clone._neighbor_positions = {
            key: list(value) for key, value in self._neighbor_positions.items()
        }
        clone._num_placement_stages_done = self._num_placement_stages_done
        clone._num_placement_stages = self._num_placement_stages
        clone._vertex_degrees = dict(self._vertex_degrees)
        return clone



class CaseData:
    """
    A test case.

    Attributes
    ----------
    database : :class:`.MoleculeDatabase
        The database to test.

    molecule : :class:`.Molecule`
        The molecule to put and get from the :attr:`.database`.

    key : :class:`object`
        The key used to retrieve :attr:`.molecule` from the database.

    """

    def __init__(self, database, molecule, key):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        database : :class:`.MoleculeDatabase`
            The database to test.

        molecule : :class:`.Molecule`
            The molecule to put and get from the `database`.

        key : :class:`object`
            The key used to retrieve `molecule` from the database.

        """

        self.database = database
        self.molecule = molecule
        self.key = key



import pathlib

import atomlite
import numpy as np

import stk

from .case_data import CaseData


def test_atomlite(molecule: CaseData) -> None:
    """Test :class:`.AtomliteDatabase`."""
    path = pathlib.Path(__file__).resolve().parent / "test.db"

    if path.exists():
        path.unlink()

    database = atomlite.Database(path)
    assert database.num_entries() == 0
    for mol, prop in zip(
        molecule.molecules,
        molecule.property_dicts,
        strict=True,
    ):
        key = stk.Smiles().get_key(mol)

        database.add_entries(
            atomlite.Entry(
                key=key,
                molecule=mol.to_atomlite(),
                properties=prop,
            )
        )

        entry = database.get_entry(key)

        assert np.allclose(
            mol.get_position_matrix(),
            mol.with_structure_from_atomlite(
                database=database,
                key=key,
            ).get_position_matrix(),
        )

        assert entry.molecule == atomlite.json_from_rdkit(mol.to_rdkit_mol())
        assert entry.molecule == mol.to_atomlite()
        assert (
            stk.Smiles().get_key(
                mol.init_from_atomlite(database=database, key=key)
            )
            == key
        )

        assert database.get_property(key, path="$.1") == prop["1"]
        assert isinstance(database.get_property(key, path="$.1"), int)
        assert database.get_property(key, path="$.3") == prop["3"]
        assert isinstance(database.get_property(key, path="$.3"), str)
        if "2" in prop:
            assert database.get_property(key=key, path="$.2") == prop["2"]
            assert isinstance(database.get_property(key, path="$.2"), dict)

    assert database.num_entries() == molecule.expected_count

    path.unlink()



import pathlib
import typing
from collections import Counter
from collections.abc import Callable, Iterator, Set
from functools import wraps
from typing import Any

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.selection.batch import Batch, BatchKey
from stk._internal.ea.selection.selectors.selector import Selector
from stk._internal.key_makers.inchi_key import InchiKey

T = typing.TypeVar("T", bound=MoleculeRecord)


class SelectMethod(typing.Protocol[T]):
    def __call__(
        self,
        population: dict[T, float],
        included_batches: Set[BatchKey] | None = None,
        excluded_batches: Set[BatchKey] | None = None,
    ) -> Iterator[Batch[T]]:
        pass


class SelectionPlotter(typing.Generic[T]):
    """
    Plots which molecule records a :class:`.Selector` selects.

    Examples:
        *Plotting Which Molecule Records Got Selected*

        .. testcode:: plotting-which-molecule-records-got-selected

            import stk

            # Make a selector.
            roulette = stk.Roulette(num_batches=10)

            # Make a population.
            population = {
                stk.MoleculeRecord(
                    topology_graph=stk.polymer.Linear(
                        building_blocks=(
                            stk.BuildingBlock(
                                smiles='BrCCBr',
                                functional_groups=[stk.BromoFactory()],
                            ),
                        ),
                        repeating_unit='A',
                        num_repeating_units=2,
                    ),
                ): i
                for i in range(100)
            }

            # Make a plotter. You do not have to assign it to a variable.
            stk.SelectionPlotter('roulette_counter', roulette)

            # Select the molecule records.
            selected = tuple(roulette.select(population))

            # There should now be a file called "roulette_counter_1.png"
            # which shows a graph of all the selected records.

            # Select records again.
            selected2 = tuple(roulette.select(population))

            # There should now be a file called "roulette_counter_2.png"
            # which shows a graph of all the selected molecules.

            # And so on every time you use "roulette.select()".

        .. testcode:: plotting-which-molecule-records-got-selected
            :hide:

            import os

            assert os.path.exists('roulette_counter_1.png')
            assert os.path.exists('roulette_counter_1.csv')
            assert os.path.exists('roulette_counter_2.png')
            assert os.path.exists('roulette_counter_2.csv')
            os.remove('roulette_counter_1.png')
            os.remove('roulette_counter_1.csv')
            os.remove('roulette_counter_2.png')
            os.remove('roulette_counter_2.csv')
    """

    def __init__(
        self,
        filename: pathlib.Path | str,
        selector: Selector[T],
        x_label: str = "Molecule: InChIKey - Fitness Value",
        record_label: Callable[
            [MoleculeRecord[Any], float], str
        ] = lambda record, fitness_value: (
            f"{InchiKey().get_key(record.get_molecule())} - {fitness_value}"
        ),
        heat_map_value: Callable[
            [MoleculeRecord[Any], float], float
        ] = lambda record, fitness_value: fitness_value,
        heat_map_label: str = "Fitness",
        order_by: Callable[
            [MoleculeRecord[Any], float], float
        ] = lambda record, fitness_value: fitness_value,
    ) -> None:
        """
        Parameters:

            filename:
                The basename of the files. This means it should not include
                file extensions.

            selector:
                The :class:`.Selector` whose selection of molecule records
                is plotted.

            x_label:
                The label use for the x axis.

            record_label:
                A function which takes a :class:`.MoleculeRecord` and its
                fitness value and returns a string, which is the label used
                for the :class:`.MoleculeRecord` on the plot.

            heat_map_value:
                A function which takes a :class:`.MoleculeRecord` and its
                fitness value and returns a value. The value is used for
                coloring the heat map used in the plot.

            heat_map_label:
                The label used for the heat map key.

            order_by:
                A function which takes a :class:`.MoleculeRecord` and its
                fitness value and returns a value. The value is used to
                sort the plotted records along the x-axis in descending
                order.
        """
        self._plots = 0
        self._filename = filename
        self._x_label = x_label
        self._record_label = record_label
        self._order_by = order_by
        self._heat_map_value = heat_map_value
        self._heat_map_label = heat_map_label
        selector.select = self._update_counter(selector.select)  # type: ignore

    def _update_counter(self, select: SelectMethod[T]) -> SelectMethod[T]:
        """
        Decorate :meth:`.Selector.select`.

        This is a decorator which makes sure that every time
        :meth:`.Selector.select` selects a :class:`.MoleculeRecord` a
        counter keeping track of selected records is updated.

        Parameters:
            select:
                The :meth:`Selector.select` method to decorate.

        Returns:
            The decorated :meth:`.Selector.select` method.

        """

        @wraps(select)
        def inner(
            population: dict[T, float],
            included_batches: Set[BatchKey] | None = None,
            excluded_batches: Set[BatchKey] | None = None,
        ) -> Iterator[Batch[T]]:
            counter = Counter({record: 0 for record in population})
            for selected in select(
                population=population,
                included_batches=included_batches,
                excluded_batches=excluded_batches,
            ):
                counter.update(selected)
                yield selected
            self._plot(population, counter)

        return inner

    def _plot(self, population: dict[T, float], counter: Counter) -> None:
        """
        Plot a selection counter.

        Parameters:
            counter:
                A counter specifying which records were selected and how
                many times.
        """
        self._plots += 1
        sns.set(style="darkgrid")
        data = []
        for record, selection_count in counter.items():
            label = self._record_label(record, population[record])
            data.append(
                pd.DataFrame(
                    data={
                        self._x_label: label,
                        "Number of Times Selected": selection_count,
                        "order": self._order_by(record, population[record]),
                        "heat_map": self._heat_map_value(
                            record,
                            population[record],
                        ),
                    },
                    index=[self._x_label],
                )
            )
        df = pd.concat(data, ignore_index=True)

        df = df.sort_values(
            ["Number of Times Selected", "order"],
            ascending=[False, False],
        )
        norm = plt.Normalize(df["heat_map"].min(), df["heat_map"].max())
        sm = plt.cm.ScalarMappable(cmap="magma_r", norm=norm)
        sm.set_array([])

        df.to_csv(f"{self._filename}_{self._plots}.csv")
        fig, ax = plt.subplots(figsize=(11.7, 8.28))
        sns.scatterplot(
            x="Number of Times Selected",
            y=self._x_label,
            hue="heat_map",
            palette="magma_r",
            data=df,
            size=[200 for _ in range(len(counter.keys()))],
            ax=ax,
        )
        ax.get_legend().remove()  # type: ignore[union-attr]
        # https://tinyurl.com/2p9drmkh
        plt.rcParams["axes.grid"] = False
        ax.figure.colorbar(sm, ax=ax).set_label(self._heat_map_label)
        plt.tight_layout()
        fig.savefig(f"{self._filename}_{self._plots}.png", dpi=fig.dpi)
        plt.close("all")



import numpy as np
import pytest
import stk

from ....case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix2(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1[C+2][C+](Br)[C+]2[C+](F)[C+2]"
                                "C2(Br)[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                ),
            ),
            smiles=(
                "F[C+]1[C+2]C23[C+2][C+]4[C+2][C+](C5=C(N=[C+]5)C56[C+"
                "2][C+]7[C+2][C+](C8=C(N=[C+]8)C89[C+2][C+]([C+2][C+]("
                "C%10=C([C+]=N%10)C%10%11[C+2][C+]([C+2][C+](C%12=C2[C"
                "+]=N%12)[C+]%10[C+](F)[C+2]%11)C2=C4[C+]=N2)[C+]8[C+]"
                "(F)[C+2]9)C2=C7[C+]=N2)[C+]5[C+](F)[C+2]6)[C+]13"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix2(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1[C+2][C+](Br)[C+]2[C+](F)[C+2]"
                                "C2(Br)[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    vertex_positions={0: np.array([2, 0, 2])},
                ),
            ),
            smiles=(
                "F[C+]1[C+2]C23[C+2][C+]4[C+2][C+](C5=C(N=[C+]5)C56[C+2"
                "][C+]7[C+2][C+](C8=C(N=[C+]8)C89[C+2][C+]([C+2][C+](C%"
                "10=C(N=[C+]%10)C%10%11[C+2][C+]([C+2][C+](C%12=C2[C+]="
                "N%12)[C+]%10[C+](F)[C+2]%11)C2=C4[C+]=N2)[C+]8[C+](F)["
                "C+2]9)C2=C7[C+]=N2)[C+]5[C+](F)[C+2]6)[C+]13"
            ),
            name=name,
        ),
    ),
)
def cage_four_plus_six_2(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



import typing
from collections.abc import Callable, Iterable
from typing import Any

import numpy as np

from .fitness_normalizer import FitnessNormalizer

T = typing.TypeVar("T")


class Add(FitnessNormalizer[T]):
    """
    Adds a number to the fitness values.

    Examples:

        *Incrementing Fitness Values by a Set of Values*

        In this example, assume that each fitness value consists of a
        :class:`tuple` of numbers, each representing a different property
        of the molecule, and each contributing to the final fitness value.
        The properties can be anything, such as  energy, number of atoms
        or diameter.

        Often, if these properties indicate a low fitness value, you
        will take their inverse. However, if these properties can have
        a value of 0, and you try to take their inverse you can end up
        dividing by 0, which is bad. To avoid this, you can add a number,
        like 1, to the fitness values before taking their inverse. This
        normalizer allows you to do this.

        Giving a concrete example

        .. testcode:: incrementing-fitness-values-by-a-set-of-values

            import stk
            import numpy as np

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record: (0, 0, 0),
            }
            normalizer = stk.Add((1, 2, 3))
            normalized_fitness_values = normalizer.normalize(fitness_values)
            assert np.all(np.equal(
                normalized_fitness_values[record],
                (1, 2, 3),
            ))

        *Selectively Normalizing Fitness Values*

        Sometimes, you only want to normalize some members of a population,
        for example if some do not have an assigned fitness value,
        because the fitness calculation failed for whatever reason.
        You can use the `filter` parameter to exclude records from the
        normalization

        .. testcode:: incrementing-fitness-values-by-a-set-of-values

            import stk
            import numpy as np

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record1 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            record2 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record1: (0, 0, 0),
                record2: None,
            }
            normalizer = stk.Add(
                number=(1, 2, 3),
                # Only normalize values which are not None.
                filter=lambda fitness_values, record:
                    fitness_values[record] is not None,
            )
            normalized_fitness_values = normalizer.normalize(
                fitness_values=fitness_values,
            )
            assert np.all(np.equal(
                normalized_fitness_values[record1],
                (1, 2, 3),
            ))
            assert normalized_fitness_values[record2] is None
    """

    def __init__(
        self,
        number: float | Iterable[float],
        filter: Callable[[dict[T, Any], T], bool] = lambda fitness_values,
        record: True,
    ) -> None:
        """
        Parameters:
            number (float | list[float]):
                The number each fitness value is increased by. Can
                be a single number or multiple numbers, depending on the
                form of the fitness value.

            filter:
                A function which returns ``True`` or ``False``. Only
                molecules which return ``True`` will have fitness values
                normalized. By default, all molecules will have fitness
                values normalized.
                The instance passed to the `fitness_values` argument of
                :meth:`.normalize` is passed as the first argument, while
                the second argument will be passed every
                :class:`.MoleculeRecord` in it, one at a time.
        """
        if not isinstance(number, int | float):
            number = tuple(number)
        self._number = number
        self._filter = filter

    def normalize(self, fitness_values: dict[T, Any]) -> dict[T, Any]:
        return {
            record: (
                np.add(self._number, fitness_value)
                if self._filter(fitness_values, record)
                else fitness_value
            )
            for record, fitness_value in fitness_values.items()
        }



class CaseData:
    """
    An :class:`.Atom` test case.

    Attributes
    ----------
    atom : :class:`.Atom`
        The atom being tested.

    id : :class:`int`
        The correct id.

    charge : :class:`int`
        The correct charge.

    atomic_number : :class:`int`
        The correct atomic number.

    """

    def __init__(self, atom, id, charge, atomic_number):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        atom : :class:`.Atom`
            The atom being tested.

        id : :class:`int`
            The correct id.

        charge : :class:`int`
            The correct charge.

        atomic_number : :class:`int`
            The correct atomic number.

        """

        self.atom = atom
        self.id = id
        self.charge = charge
        self.atomic_number = atomic_number



"""
Collapser
=========

"""

import mchammer as mch

from .optimizer import Optimizer
from .utilities import get_long_bond_ids, get_mch_bonds, get_subunits


class Collapser(Optimizer):
    """
    Performs rigid-body collapse of molecules.

    Examples:

        *Structure Optimization*

        Using :class:`.Collapser` will lead to
        :class:`.ConstructedMolecule` structures without long bonds.

        .. testcode:: structure-optimization

            import stk

            bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
            bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])

            polymer = stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(bb1, bb2),
                    repeating_unit='AB',
                    num_repeating_units=2,
                    optimizer=stk.Collapser(),
                ),
            )

        Optimisation with :mod:`stk` simply collects the final position
        matrix. The optimisation's trajectory can be output using the
        :mod:`MCHammer` implementation if required by the
        user [#mchammer]_.

        The open-source optimization code :mod:`MCHammer` specializes in
        the `collapsing` of molecules with long bonds like those
        constructed by :mod:`stk`. This code is entirely nonphysical and
        is, therefore, completely general to any chemistry.

    References:

        .. [#mchammer] https://github.com/andrewtarzia/MCHammer

    """

    def __init__(
        self,
        step_size=0.1,
        distance_threshold=1.5,
        scale_steps=True,
    ):
        """
        Initialize an instance of :class:`.Collapser`.

        Parameters
        ----------
        step_size : :class:`float`, optional
            The relative size of the step to take during collapse in
            Angstrom.

        distance_threshold : :class:`float`, optional
            Distance between distinct building blocks to use as
            threshold for halting collapse in Angstrom.

        scale_steps : :class:`bool`, optional
            Whether to scale the step of each distinct building block
            by its relative distance from the molecules centroid.

        """

        self._optimizer = mch.Collapser(
            step_size=step_size,
            distance_threshold=distance_threshold,
            scale_steps=scale_steps,
        )

    def optimize(self, state):
        # Define MCHammer molecule to optimize.
        mch_mol = mch.Molecule(
            atoms=(
                mch.Atom(
                    id=atom.get_id(),
                    element_string=atom.__class__.__name__,
                )
                for atom in state.get_atoms()
            ),
            bonds=get_mch_bonds(state),
            position_matrix=state.get_position_matrix(),
        )

        # Run optimization.
        mch_mol, result = self._optimizer.get_result(
            mol=mch_mol,
            bond_pair_ids=tuple(get_long_bond_ids(state)),
            subunits=get_subunits(state),
        )
        return state.with_position_matrix(
            position_matrix=mch_mol.get_position_matrix()
        )



import pytest
import stk

from ....case_data import CaseData


@pytest.fixture
def cof1(cls, id, position, aligner_edge, cell):
    return CaseData(
        vertex=cls(
            id=id,
            position=position,
            aligner_edge=aligner_edge,
            cell=cell,
        ),
        id=id,
        cell=cell,
        position=position,
    )


@pytest.fixture(
    params=(
        stk.cof.LinearVertex,
        stk.cof.NonLinearVertex,
        stk.cof.UnaligningVertex,
    ),
)
def cls(request):
    return request.param



"""
Host Guest Complex Vertices
===========================

"""

import numpy as np

from stk._internal.topology_graphs.vertex import Vertex


class HostVertex(Vertex):
    """
    Places the host in a :class:`.Complex`.

    """

    def place_building_block(self, building_block, edges):
        return building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        return {}


class GuestVertex(Vertex):
    """
    Places the guest in a :class:`.Complex`.

    """

    def __init__(self, id, position, start, target):
        """
        Initialize a :class:`.GuestVertex` instance.

        Parameters
        ----------
        id : :class:`int`
            The id of the vertex.

        position : :class:`tuple` of :class:`float`
            The position of the vertex.

        start : :class:`tuple` of :class:`float`
            A vector which defines the start of the rotation applied
            to the molecules placed by the vertex.

        target : :class:`tuple` of :class:`float`
            A vector which defines the end of the rotation applied
            to the molecules placed by the vertex.

        """

        self._start = np.array(start, dtype=np.float64)
        self._target = np.array(target, dtype=np.float64)
        super().__init__(id, position)

    def clone(self):
        clone = super().clone()
        clone._start = np.array(self._start)
        clone._target = np.array(self._target)
        return clone

    def place_building_block(self, building_block, edges):
        return (
            building_block.with_centroid(
                position=self._position,
                atom_ids=building_block.get_placer_ids(),
            )
            .with_rotation_between_vectors(
                start=self._start,
                target=self._target,
                origin=self._position,
            )
            .get_position_matrix()
        )

    def map_functional_groups_to_edges(self, building_block, edges):
        return {}



import pytest
import stk

from ....case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.SixPlusTwelve(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1C2(Br)[C+]=N[C+]2[C+](Br)[C+]("
                                "Br)[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                ),
            ),
            smiles=(
                "[C+]1=NC2=C1[C+]1[C+2][C+]3C4=C(N=[C+]4)[C+]4[C+]5[C+"
                "2][C+]6C7=C(N=[C+]7)[C+]7[C+]8[C+2][C+]9C%10=C(N=[C+]"
                "%10)[C+]%10[C+2][C+]%11C%12=C(N=[C+]%12)[C+]%12[C+2]["
                "C+](C%13=C([C+]=N%13)C9%13[C+]=N[C+]7%13)[C+](C7=C([C"
                "+]=N7)C67[C+]=N[C+]47)[C+]4N=[C+]C%124C4=C([C+]=N4)C3"
                "4[C+]=N[C+]4[C+]1C1=C(N=[C+]1)[C+]%11[C+]1N=[C+]C%101"
                "C1=C(N=[C+]1)[C+]1[C+2][C+]2[C+](C2=C5[C+]=N2)[C+]2N="
                "[C+]C12C1=C8[C+]=N1"
            ),
            name=name,
        ),
    ),
)
def cage_six_plus_twelve(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



from __future__ import annotations

import typing

from stk._internal.atom import Atom

_T = typing.TypeVar("_T", bound="FunctionalGroup")


class FunctionalGroup:
    """
    An abstract base class for functional groups.

    It is used to give access to atoms of :class:`.BuildingBlock`
    molecules which are modified during
    :class:`.ConstructedMolecule` construction, as well as specify
    which atoms of the building block should be used for positioning.

    *Should I use* :meth:`.with_ids` *or* :meth:`.with_atoms` *?*

    That depends on your use case, however, it is generally better to
    default to :meth:`.with_ids` unless you need to actually change
    the atoms held by the functional group. This is because
    :meth:`.with_ids` preserves the most-derived type of the functional
    group, while :meth:`.with_atoms` does not. To give an example

    .. testcode:: with-atoms-vs-with-ids

        import stk

        bromo = stk.Bromo(
            bromine=stk.Br(0),
            atom=stk.C(1),
            bonders=(stk.C(1), ),
            deleters=(stk.Br(0), ),
        )

        bromo2 = bromo.with_ids({
            0: 10,
            1: 100,
        })
        # bromo2 is still a Bromo functional group.
        assert isinstance(bromo2, stk.Bromo)

        not_bromo = bromo.with_atoms({
            0: stk.Br(10),
            1: stk.C(100),
        })
        # not_bromo is not a Bromo functional gorup.
        assert not isinstance(not_bromo, stk.Bromo)
        # However, it is still an instance of FunctionalGroup,
        assert isinstance(not_bromo, stk.FunctionalGroup)
        # and of GenericFunctionalGroup
        assert isinstance(not_bromo, stk.GenericFunctionalGroup)

    The reason that :meth:`.with_atoms` does not produce a
    :class:`.Bromo` instance is to avoid the following pitfall

    .. code-block:: python

        pitfall = bromo.with_atoms({
            0: stk.F(10),
            1: stk.C(100),
        })
        # If with_atoms() returned a Bromo instance then you could
        # call get_bromine() on it, but it would hold a F atom!
        this_is_a_fluorine = pitfall.get_bromine()

    *Why would I want to implement a new subclass?*

    The most common reason you would want to implement a new
    :class:`.FunctionalGroup` subclass, is because you want to
    customize the construction of a :class:`.ConstructedMolecule`.
    Specifically, you want to modify a specific set of atoms in a
    :class:`.BuildingBlock` when doing construction, and you want
    to modify them in a specific way. You will usually accompany the
    creation of the new :class:`.FunctionalGroup` subclass with the
    creation of a new :class:`.Reaction` subclass, which will perform
    the custom modification on the atoms held by your new
    :class:`.FunctionalGroup`
    subclass. Finally, a new :class:`.ReactionFactory` will also be
    created, so that your :class:`.Reaction` subclass instances
    actually get made during construction. Finally, you will pass an
    instance of your :class:`.ReactionFactory` subclass to the
    chosen :class:`.TopologyGraph` you want to make, for example
    :class:`~.polymer.linear.Linear`, and your custom
    modification will take place.

    See Also:

        :mod:`.functional_group_factory`
            Used for automated creation of :class:`.FunctionalGroup`
            instances.

    Notes:

        You might notice that some of the methods of this abstract base
        class are implemented. This is purely for convenience when
        implementing subclasses. The implemented public methods are
        simply default implementations, which can be safely ignored
        or overridden, when implementing subclasses. Any private
        methods are implementation details of these default
        implementations.

    Examples:

        *Subclass Implementation*

        The source code of the subclasses, listed in
        :mod:`.functional_group`, can serve as good examples.

        *Changing the Atoms of a Functional Group*

        You want to substitute the atoms in the functional group for
        other atoms. You can do this by using :meth:`.with_atoms` to
        create a clone of the functional group, which holds the
        replacement atoms

        .. testcode:: changing-the-atoms-of-a-functional-group

            import stk

            c, n, h1, h2 = stk.C(0), stk.N(1), stk.H(2), stk.H(3)
            amine = stk.PrimaryAmino(
                nitrogen=n,
                hydrogen1=h1,
                hydrogen2=h2,
                atom=c,
                bonders=(n, ),
                deleters=(h1, h2),
            )

            n20 = stk.N(20)
            h100 = stk.H(100)

            # amine_clone is a clone of amine, except that instead of
            # holding n, amine_clone holds n20, and instead of holding
            # h1  amine_clone holds h100. amine_clone continues to hold
            # h2.
            amine_clone = amine.with_atoms({
                n.get_id(): n20,
                h1.get_id(): h100,
            })

        .. testcode:: changing-the-atoms-of-a-functional-group
            :hide:

            _atoms = set(amine.get_atom_ids())
            assert n.get_id() in _atoms
            assert n20.get_id() not in _atoms
            assert h1.get_id() in _atoms
            assert h100.get_id()  not in _atoms

            _clone_atoms = set(amine_clone.get_atom_ids())
            assert n.get_id() not in _clone_atoms
            assert n20.get_id() in _clone_atoms
            assert h1.get_id() not in _clone_atoms
            assert h100.get_id() in _clone_atoms

            _bonders = set(amine.get_bonder_ids())
            assert n.get_id() in _bonders
            assert n20.get_id() not in _bonders

            _clone_bonders = set(amine_clone.get_bonder_ids())
            assert n.get_id() not in _clone_bonders
            assert n20.get_id() in _clone_bonders

            _deleters = set(amine.get_deleter_ids())
            assert h1.get_id() in _deleters
            assert h2.get_id() in _deleters
            assert h100.get_id() not in _deleters

            _clone_deleters = set(amine_clone.get_deleter_ids())
            assert h1.get_id() not in _clone_deleters
            assert h2.get_id() in _clone_deleters
            assert h100.get_id() in _clone_deleters

    """

    def __init__(
        self,
        atoms: tuple[Atom, ...],
        placers: tuple[Atom, ...],
        core_atoms: tuple[Atom, ...],
    ) -> None:
        """
        Initialize a :class:`.FunctionalGroup`.

        Parameters:

            atoms:
                The atoms in the functional group.

            placers:
                The atoms used to calculate the position of the
                functional group.

            core_atoms:
                The atoms of the functional group which also form the
                core of the :class:`.BuildingBlock`. See
                :meth:`.BuildingBlock.get_core_atom_ids`.

        """

        self._atoms = atoms
        self._placers = placers
        self._core_atoms = core_atoms

    def get_atoms(self) -> typing.Iterator[Atom]:
        """
        Yield all the atoms in the functional group.

        Yields:
            An atom in the functional group.

        """

        yield from self._atoms

    def get_atom_ids(self) -> typing.Iterator[int]:
        """
        Yield the ids of all atoms in the functional group.

        Yields:
            The id of an :class:`.Atom`.

        """

        yield from (a.get_id() for a in self._atoms)

    def get_placer_ids(self) -> typing.Iterator[int]:
        """
        Yield the ids of *placer* atoms.

        *Placer* atoms are those, which should be used to calculate
        the position of the functional group.

        Yields:
            The id of an :class:`.Atom`.

        """

        yield from (a.get_id() for a in self._placers)

    def get_core_atom_ids(self) -> typing.Iterator[int]:
        """
        Yield the ids of core atoms held by the functional group.

        Yields:
            The id of an :class:`.Atom`.

        See Also:
            :meth:`.BuildingBlock.get_core_atom_ids`

        """

        yield from (a.get_id() for a in self._core_atoms)

    def with_atoms(self, atom_map: dict[int, Atom]) -> FunctionalGroup:
        """
        Return a clone holding different atoms.

        Parameters:
            atom_map:
                Maps the id of an atom in the functional group to the
                new atom the clone should hold. If the id of an atom in
                the functional group is not found in `atom_map`, the
                atom will not be replaced in the clone.

        Returns:
            The clone.

        """

        # The clone needs to be downcasted.
        return FunctionalGroup(
            atoms=tuple(atom_map.get(a.get_id(), a) for a in self._atoms),
            placers=tuple(atom_map.get(a.get_id(), a) for a in self._placers),
            core_atoms=tuple(
                atom_map.get(a.get_id(), a) for a in self._core_atoms
            ),
        )

    def _with_ids(self, id_map: dict[int, int]) -> FunctionalGroup:
        self._atoms = tuple(
            atom.with_id(
                id=id_map.get(
                    atom.get_id(),
                    atom.get_id(),
                ),
            )
            for atom in self._atoms
        )
        self._placers = tuple(
            atom.with_id(
                id=id_map.get(
                    atom.get_id(),
                    atom.get_id(),
                ),
            )
            for atom in self._placers
        )
        self._core_atoms = tuple(
            atom.with_id(
                id=id_map.get(
                    atom.get_id(),
                    atom.get_id(),
                ),
            )
            for atom in self._core_atoms
        )
        return self

    def with_ids(
        self,
        id_map: dict[int, int],
    ) -> FunctionalGroup:
        """
        Return a clone holding different atom ids.

        Parameters:
            id_map:
                Maps the id of an atom in the functional group to the
                new id the clone should hold. If the id of an atom in
                the functional group is not found in `id_map`, the
                atom will not be replaced in the clone.

        Returns:
            The clone.

        """

        return self.clone()._with_ids(id_map)

    def _clone(self: _T) -> _T:
        """
        Return a clone.

        Notes:

            This method exists so that subclasses can use it and have
            correct type information, because the type variable
            ``_T`` is used. However, I do not want the docs to show
            ``_T`` I want them to show a proper class, such as
            :class:`.Oxygen`. So this method exists for re-use and
            type checking, while the public methods exist to provide
            a readable type for the docs.

        Returns:
            A clone.

        """

        clone = self.__class__.__new__(self.__class__)
        FunctionalGroup.__init__(
            self=clone,
            atoms=self._atoms,
            placers=self._placers,
            core_atoms=self._core_atoms,
        )
        return clone

    def clone(self) -> FunctionalGroup:
        """
        Return a clone.

        Returns:
            A clone.

        """

        return self._clone()

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"atoms={self._atoms}, "
            f"placers={self._placers}, "
            f"core_atoms={self._core_atoms}"
            ")"
        )



"""
Eight Plus Twelve
=================

"""

import numpy as np

from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import LinearVertex, NonLinearVertex


class EightPlusTwelve(Cage):
    """
    Represents a cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='NC1CCCCC1N',
            functional_groups=[stk.PrimaryAminoFactory()],
        )
        bb2 = stk.BuildingBlock(
            smiles='O=Cc1cc(C=O)cc(C=O)c1',
            functional_groups=[stk.AldehydeFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.EightPlusTwelve((bb1, bb2)),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.MCHammer` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='NC1CCCCC1N',
            functional_groups=[stk.PrimaryAminoFactory()],
        )
        bb2 = stk.BuildingBlock(
            smiles='O=Cc1cc(C=O)cc(C=O)c1',
            functional_groups=[stk.AldehydeFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.EightPlusTwelve(
                building_blocks=(bb1, bb2),
                optimizer=stk.MCHammer(num_steps=1000),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    Nonlinear building blocks with three functional groups are
    required for this topology.

    Linear building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 3-functional groups: 0 to 7
        | 2-functional groups: 8 to 19

    See :class:`.Cage` for more details and examples.

    """

    _non_linears = (
        NonLinearVertex(0, np.array([-1, 1, -1])),
        NonLinearVertex(1, np.array([-1, -1, -1])),
        NonLinearVertex(2, np.array([1, 1, -1])),
        NonLinearVertex(3, np.array([1, -1, -1])),
        NonLinearVertex(4, np.array([-1, 1, 1])),
        NonLinearVertex(5, np.array([-1, -1, 1])),
        NonLinearVertex(6, np.array([1, 1, 1])),
        NonLinearVertex(7, np.array([1, -1, 1])),
    )

    _vertex_prototypes = (
        *_non_linears,
        LinearVertex.init_at_center(
            id=8,
            vertices=(_non_linears[0], _non_linears[2]),
        ),
        LinearVertex.init_at_center(
            id=9,
            vertices=(_non_linears[0], _non_linears[1]),
        ),
        LinearVertex.init_at_center(
            id=10,
            vertices=(_non_linears[1], _non_linears[3]),
        ),
        LinearVertex.init_at_center(
            id=11,
            vertices=(_non_linears[2], _non_linears[3]),
        ),
        LinearVertex.init_at_center(
            id=12,
            vertices=(_non_linears[4], _non_linears[6]),
        ),
        LinearVertex.init_at_center(
            id=13,
            vertices=(_non_linears[4], _non_linears[5]),
        ),
        LinearVertex.init_at_center(
            id=14,
            vertices=(_non_linears[5], _non_linears[7]),
        ),
        LinearVertex.init_at_center(
            id=15,
            vertices=(_non_linears[6], _non_linears[7]),
        ),
        LinearVertex.init_at_center(
            id=16,
            vertices=(_non_linears[0], _non_linears[4]),
        ),
        LinearVertex.init_at_center(
            id=17,
            vertices=(_non_linears[1], _non_linears[5]),
        ),
        LinearVertex.init_at_center(
            id=18,
            vertices=(_non_linears[2], _non_linears[6]),
        ),
        LinearVertex.init_at_center(
            id=19,
            vertices=(_non_linears[3], _non_linears[7]),
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[8], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[8], _vertex_prototypes[2]),
        Edge(2, _vertex_prototypes[9], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[9], _vertex_prototypes[1]),
        Edge(4, _vertex_prototypes[10], _vertex_prototypes[1]),
        Edge(5, _vertex_prototypes[10], _vertex_prototypes[3]),
        Edge(6, _vertex_prototypes[11], _vertex_prototypes[2]),
        Edge(7, _vertex_prototypes[11], _vertex_prototypes[3]),
        Edge(8, _vertex_prototypes[12], _vertex_prototypes[4]),
        Edge(9, _vertex_prototypes[12], _vertex_prototypes[6]),
        Edge(10, _vertex_prototypes[13], _vertex_prototypes[4]),
        Edge(11, _vertex_prototypes[13], _vertex_prototypes[5]),
        Edge(12, _vertex_prototypes[14], _vertex_prototypes[5]),
        Edge(13, _vertex_prototypes[14], _vertex_prototypes[7]),
        Edge(14, _vertex_prototypes[15], _vertex_prototypes[6]),
        Edge(15, _vertex_prototypes[15], _vertex_prototypes[7]),
        Edge(16, _vertex_prototypes[16], _vertex_prototypes[0]),
        Edge(17, _vertex_prototypes[16], _vertex_prototypes[4]),
        Edge(18, _vertex_prototypes[17], _vertex_prototypes[1]),
        Edge(19, _vertex_prototypes[17], _vertex_prototypes[5]),
        Edge(20, _vertex_prototypes[18], _vertex_prototypes[2]),
        Edge(21, _vertex_prototypes[18], _vertex_prototypes[6]),
        Edge(22, _vertex_prototypes[19], _vertex_prototypes[3]),
        Edge(23, _vertex_prototypes[19], _vertex_prototypes[7]),
    )

    _num_windows = 6
    _num_window_types = 1



import numpy as np

from ...utilities import is_clone


def test_with_centroid(molecule, get_atom_ids, centroid):
    """
    Test :meth:`.Molecule.with_centroid`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `atom_ids` parameter for :meth:`.Molecule.with_centroid`. This
        allows the testing of different values of this parameter.

    centroid : :class:`numpy.ndarray`
        The desired centroid of the clone molecule.

    Returns
    -------
    None : :class:`NoneType`

    """

    # Save a copy of the position matrix, to ensure the original
    # molecule is not modified by the test, because it is meant to be
    # immutable.
    position_matrix = molecule.get_position_matrix()
    _test_with_centroid(molecule, get_atom_ids, centroid)
    assert np.all(
        np.equal(
            position_matrix,
            molecule.get_position_matrix(),
        )
    )


def _test_with_centroid(molecule, get_atom_ids, centroid):
    """
    Test :meth:`.Molecule.with_centroid`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `atom_ids` parameter for :meth:`.Molecule.with_centroid`. This
        allows the testing of different values of this parameter.

    centroid : :class:`numpy.ndarray`
        The desired centroid of the clone molecule.

    Returns
    -------
    None : :class:`NoneType`

    """

    new = molecule.with_centroid(
        position=centroid,
        atom_ids=get_atom_ids(molecule),
    )
    is_clone(new, molecule)
    assert np.allclose(
        a=centroid,
        b=new.get_centroid(atom_ids=get_atom_ids(molecule)),
        atol=1e-8,
    )



import numpy as np
import pytest
import stk
from pytest_lazyfixture import lazy_fixture

# Fixtures need to be visible for lazy_fixture() calls.
from .fixtures import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("flip"),
        lazy_fixture("no_flip"),
    ),
)
def case_data(request):
    return request.param


@pytest.fixture(
    params=(
        lambda: stk.BuildingBlock(
            smiles="BrCCNCBr",
            functional_groups=[stk.BromoFactory()],
        ),
    ),
)
def building_block_2(request) -> stk.BuildingBlock:
    """
    A :class:`.BuildingBlock` with 2 functional groups.

    """

    return request.param()


@pytest.fixture(
    params=(
        np.pi / 2,
        0,
        np.pi,
    ),
)
def angle(request):
    return request.param



import itertools as it

import numpy as np


def test_get_reaction(case_data):
    """
    Test :class:`.ReactionFactory.get_reaction`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        The test case. Holds the factory, and the expected reaction
        result.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_reaction(
        factory=case_data.factory,
        construction_state=case_data.construction_state,
        edge_group=case_data.edge_group,
        reaction_result=case_data.reaction_result,
    )


def _test_get_reaction(
    factory,
    construction_state,
    edge_group,
    reaction_result,
):
    """
    Test :class:`.ReactionFactory.get_reaction`.

    Parameters
    ----------
    factory : :class:`.ReactionFactory`
        The factory to test.

    construction_state : :class:`.ConstructionState`
        The construction state to pass to the factory.

    edge_group : :class:`.EdgeGroup`
        The edge group to pass to the factory.

    reaction_result : :class:`.ReactionResult`
        The expected result of the reaction returned by
        `factory`.

    Returns
    -------
    None : :class:`NoneType`

    """

    reaction = factory.get_reaction(
        construction_state=construction_state,
        edge_group=edge_group,
    )
    is_same_result(reaction.get_result(), reaction_result)


def is_same_result(result1, result2):
    are_same_bonds(result1.get_new_bonds(), result2.get_new_bonds())
    are_same_new_atoms(
        new_atoms1=result1.get_new_atoms(),
        new_atoms2=result2.get_new_atoms(),
    )
    are_same_atoms(
        atoms1=result1.get_deleted_atoms(),
        atoms2=result2.get_deleted_atoms(),
    )
    are_same_bonds(
        result1.get_deleted_bonds(),
        result2.get_deleted_bonds(),
    )


def are_same_new_atoms(new_atoms1, new_atoms2):
    for (atom1, position1), (atom2, position2) in it.zip_longest(
        new_atoms1,
        new_atoms2,
    ):
        is_same_atom(atom1, atom2)
        assert np.allclose(position1, position2, atol=1e-13)


def are_same_atoms(atoms1, atoms2):
    for atom1, atom2 in it.zip_longest(atoms1, atoms2):
        is_same_atom(atom1, atom2)


def is_same_atom(atom1, atom2):
    assert atom1.__class__ is atom2.__class__
    assert atom1.get_id() == atom2.get_id()
    assert atom1.get_charge() == atom2.get_charge()


def are_same_bonds(bonds1, bonds2):
    bonds1 = sorted(bonds1, key=atom_ids)
    bonds2 = sorted(bonds2, key=atom_ids)
    for bond1, bond2 in it.zip_longest(bonds1, bonds2):
        is_same_bond(bond1, bond2)


def atom_ids(bond):
    return bond.get_atom1().get_id(), bond.get_atom2().get_id()


def is_same_bond(bond1, bond2):
    are_same_atoms(
        atoms1=sorted(
            (bond1.get_atom1(), bond1.get_atom2()),
            key=lambda atom: atom.get_id(),
        ),
        atoms2=sorted(
            (bond2.get_atom1(), bond2.get_atom2()),
            key=lambda atom: atom.get_id(),
        ),
    )
    assert bond1.get_periodicity() == bond2.get_periodicity()



class CaseData:
    """
    A :class:`.TopologyGraph` test case.

    Attributes
    ----------
    topology_graph : :class:`.TopologyGraph`
        The topology graph to test.

    cell : :class:`tuple` of :class:`np.array` or :class:`NoneType`
        Tuple of cell lattice vectors (shape: (3,)) in Angstrom.
        `None` if topology graph is not periodic.

    """

    def __init__(self, topology_graph, cell):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        topology_graph : :class:`.TopologyGraph`
            The topology graph to test.

        cell : :class:`tuple` of :class:`np.array` or :class:`NoneType`
            Tuple of cell lattice vectors (shape: (3,)) in Angstrom.
            `None` if topology graph is not periodic.

        """

        self.topology_graph = topology_graph
        self.cell = cell



import stk

from .case_data import CaseData


def test_to_string(case_data: CaseData) -> None:
    """
    Test writing of molecule to a string.

    Parameters:
        case_data:
            A test case.

    """

    _test_to_string(
        molecule=case_data.molecule,
        writer=case_data.writer,
        string=case_data.string,
    )


def _test_to_string(
    molecule: stk.Molecule,
    writer: stk.XyzWriter,
    string: str,
) -> None:
    """
    Test that the written string matches expected string.

    Parameters:
        molecule:
            Molecule to test.

        writer:
            The writer to test.

        string:
            The expected output string.

    """

    test_string = writer.to_string(molecule)

    assert test_string == string



import stk

from .case_data import CaseData


def test_to_string(case_data: CaseData) -> None:
    """
    Test writing of molecule to a string.

    Parameters:
        case_data:
            A test case.

    """

    _test_to_string(
        molecule=case_data.molecule,
        writer=case_data.writer,
        string=case_data.string,
        periodic_info=case_data.periodic_info,
    )


def _test_to_string(
    molecule: stk.Molecule,
    writer: stk.TurbomoleWriter,
    string: str,
    periodic_info: stk.PeriodicInfo | None = None,
) -> None:
    """
    Test that the written string matches expected string.

    Parameters:
        molecule:
            Molecule to test.

        writer:
            The writer to test.

        string:
            The expected output string.

        periodic_info:
            Periodic information.

    """

    test_string = writer.to_string(
        molecule=molecule,
        periodic_info=periodic_info,
    )

    assert test_string == string



import itertools as it

import pytest
import stk
from pytest_lazyfixture import lazy_fixture
from collections import abc

# Fixtures need be visible for lazy_fixture() calls.
from .fixtures import *  # noqa

from .case_data import CaseData


@pytest.fixture(
    params=(
        lazy_fixture("default_init"),
        lazy_fixture("init_from_file"),
        lazy_fixture("init"),
        lazy_fixture("init_from_molecule"),
        lazy_fixture("init_from_rdkit_mol"),
    ),
)
def case_data(request) -> CaseData:
    """
    A :class:`.CaseData` instance.

    """

    return request.param


@pytest.fixture
def building_block(case_data: CaseData) -> stk.BuildingBlock:
    """
    A :class:`.BuildingBlock` instance.

    """

    return case_data.building_block


@pytest.fixture(
    params=(
        lambda molecule: stk.BromoFactory().get_functional_groups(molecule),
        lambda molecule: stk.PrimaryAminoFactory().get_functional_groups(
            molecule
        ),
        lambda molecule: it.chain(
            stk.PrimaryAminoFactory().get_functional_groups(molecule),
            stk.BromoFactory().get_functional_groups(molecule),
        ),
    )
)
def get_functional_groups(
    request: pytest.FixtureRequest,
) -> abc.Iterable[stk.FunctionalGroup]:
    """
    Get the functional groups of a `molecule`.

    Parameters:
        molecule:
            The molecule whose functional groups should be gotten.

    Returns:
        A functional group of `molecule`.

    """

    return request.param



import itertools as it


def test_get_placer_ids(case_data):
    """
    Test :meth:`.FunctionalGroup.get_placer_ids`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the functional group to test and the atoms
        holding the correct ids.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_placer_ids(
        functional_group=case_data.functional_group,
        placers=case_data.placers,
    )


def _test_get_placer_ids(functional_group, placers):
    """
    Test :meth:`.FunctionalGroup.get_placer_ids`.

    Parameters
    ----------
    functional_group : :class:`.FunctionalGroup`
        The functional group to test.

    placers : :class:`tuple` of :class:`.Atom`
        The atoms with correct ids.

    Returns
    -------
    None : :class:`NoneType`

    """

    for id_, placer in it.zip_longest(
        functional_group.get_placer_ids(),
        placers,
    ):
        assert id_ == placer.get_id()



import pytest
from pytest_lazyfixture import lazy_fixture

from ...case_data import CaseData

# Fixtures need to be visible for lazy_fixture() calls.
from .linear import *  # noqa


@pytest.fixture(
    params=(lazy_fixture("polymer_linear"),),
)
def polymer(request: pytest.FixtureRequest) -> CaseData:
    return request.param



import logging
import os
import pathlib
import typing
from collections.abc import Collection, Iterable, Iterator
from functools import partial

import atomlite
import numpy as np
import rdkit.Chem.AllChem as rdkit
import vabene

from stk._internal.atom import Atom
from stk._internal.bond import Bond
from stk._internal.functional_group_factories.functional_group_factory import (
    FunctionalGroupFactory,
)
from stk._internal.functional_groups.functional_group import FunctionalGroup
from stk._internal.molecule import Molecule
from stk._internal.utilities.utilities import flatten, remake

logger = logging.getLogger(__name__)


class BuildingBlock(Molecule):
    """
    Represents a building block of a :class:`.ConstructedMolecule`.

    A :class:`BuildingBlock` can represent either an entire molecule or
    a molecular fragments used to construct a
    :class:`.ConstructedMolecule`. The building block uses
    :class:`.FunctionalGroup` instances to identify which atoms are
    modified during construction.

    See Also:
        * :class:`.Atom`: Represents atoms of a building block.
        * :class:`.Bond`: Represents bonds of a building block.

    """

    # Maps file extensions to functions which can be used to
    # create an rdkit molecule from that file type.
    _init_funcs = {
        ".mol": partial(rdkit.MolFromMolFile, sanitize=False, removeHs=False),
        ".sdf": partial(rdkit.MolFromMolFile, sanitize=False, removeHs=False),
    }

    _placer_ids: frozenset[int]
    _core_ids: frozenset[int]

    def __init__(
        self,
        smiles: str,
        functional_groups: (
            FunctionalGroup
            | FunctionalGroupFactory
            | Iterable[FunctionalGroup | FunctionalGroupFactory]
        ) = (),
        placer_ids: Iterable[int] | None = None,
        position_matrix: np.ndarray | None = None,
    ) -> None:
        """
        Parameters:

            smiles:
                A SMILES string of the molecule.

            functional_groups (FunctionalGroup \
| FunctionalGroupFactory \
| list[FunctionalGroup | FunctionalGroupFactory]):
                :class:`.FunctionalGroup` instances added to the
                building block and :class:`.FunctionalGroupFactory`
                instances used to create :class:`.FunctionalGroup`
                instances added to the building block.
                :class:`.FunctionalGroup` instances are used to
                identify which atoms are modified during
                :class:`.ConstructedMolecule` construction.

            placer_ids (list[int]):
                The ids of *placer* atoms. These are the atoms which
                should be used for calculating the position of the
                building block. Depending on the values passed to
                `placer_ids`, and the functional groups in the building
                block, different *placer* ids will be used by the
                building block.

                #. `placer_ids` is passed to the initializer: the
                   passed *placer* ids will be used by the building
                   block.

                #. `placer_ids` is ``None`` and the building block has
                   functional groups: The *placer* ids of the
                   functional groups will be used as the *placer* ids
                   of the building block.

                #. `placer_ids` is ``None`` and `functional_groups` is
                   empty. All atoms of the molecule will be used for
                   *placer* ids.

            position_matrix:
                The position matrix the building block should use. If
                ``None``, :func:`rdkit.ETKDGv2` will be used to
                calculate it.

        Raises:

            :class:`RuntimeError`
                If embedding the molecule fails.

        """

        molecule = rdkit.AddHs(rdkit.MolFromSmiles(smiles))
        if position_matrix is None:
            params = rdkit.ETKDGv2()
            random_seed = 4
            params.randomSeed = random_seed
            if rdkit.EmbedMolecule(molecule, params) == -1:
                raise RuntimeError(
                    f"Embedding with seed value of {random_seed} failed."
                )
            rdkit.Kekulize(molecule)
        else:
            # Make sure the position matrix always holds floats.
            position_matrix = np.array(
                position_matrix,
                dtype=np.float64,
            )
            conformer = rdkit.Conformer(molecule.GetNumAtoms())
            for atom_id, position in enumerate(position_matrix):
                conformer.SetAtomPosition(atom_id, position)
            molecule.AddConformer(conformer)

        self._init_from_rdkit_mol(
            molecule=molecule,
            functional_groups=functional_groups,
            placer_ids=placer_ids,
        )

    @classmethod
    def init_from_molecule(
        cls,
        molecule: Molecule,
        functional_groups: (
            FunctionalGroup
            | FunctionalGroupFactory
            | Iterable[FunctionalGroup | FunctionalGroupFactory]
        ) = (),
        placer_ids: Iterable[int] | None = None,
    ) -> typing.Self:
        """
        Initialize from a :class:`.Molecule`.

        Parameters:

            molecule:
                The molecule to initialize from.

            functional_groups (FunctionalGroup \
| FunctionalGroupFactory \
| list[FunctionalGroup | FunctionalGroupFactory]):
                :class:`.FunctionalGroup` instances added to the
                building block and :class:`.FunctionalGroupFactory`
                instances used to create :class:`.FunctionalGroup`
                instances added to the building block.
                :class:`.FunctionalGroup` instances are used to
                identify which atoms are modified during
                :class:`.ConstructedMolecule` construction.

            placer_ids (list[int]):
                The ids of *placer* atoms. These are the atoms which
                should be used for calculating the position of the
                building block. Depending on the values passed to
                `placer_ids`, and the functional groups in the
                building block, different *placer* ids will be used by
                the building block.

                #. `placer_ids` is passed to the initializer: the
                   passed *placer* ids will be used by the building
                   block.

                #. `placer_ids` is ``None`` and the building block has
                   functional groups: The *placer* ids of the
                   functional groups will be used as the *placer* ids
                   of the building block.

                #. `placer_ids` is ``None`` and `functional_groups` is
                   empty. All atoms of the molecule will be used for
                   *placer* ids.

        Returns:
            BuildingBlock: The building block. It will have the same
            atoms, bonds and atomic positions as `molecule`.

        """

        return cls.init(
            atoms=tuple(molecule.get_atoms()),
            bonds=tuple(molecule.get_bonds()),
            position_matrix=molecule.get_position_matrix(),
            functional_groups=functional_groups,
            placer_ids=placer_ids,
        )

    @classmethod
    def init_from_vabene_molecule(
        cls,
        molecule: vabene.Molecule,
        functional_groups: (
            FunctionalGroup
            | FunctionalGroupFactory
            | Iterable[FunctionalGroup | FunctionalGroupFactory]
        ) = (),
        placer_ids: Iterable[int] | None = None,
        position_matrix: np.ndarray | None = None,
    ) -> typing.Self:
        """
        Initialize from a :mod:`vabene.Molecule`.

        Notes:

            The molecule is given 3D coordinates with
            :func:`rdkit.ETKDGv2()`.

        Parameters:

            molecule:
                The :class:`vabene.Molecule` from which to initialize.

            functional_groups (FunctionalGroup \
| FunctionalGroupFactory \
| list[FunctionalGroup | FunctionalGroupFactory]):
                :class:`.FunctionalGroup` instances added to the
                building block and :class:`.FunctionalGroupFactory`
                instances used to create :class:`.FunctionalGroup`
                instances added to the building block.
                :class:`.FunctionalGroup` instances are used to
                identify which atoms are modified during
                :class:`.ConstructedMolecule` construction.

            placer_ids (list[int]):
                The ids of *placer* atoms. These are the atoms which
                should be used for calculating the position of the
                building block. Depending on the values passed to
                `placer_ids`, and the functional groups in the building
                block, different *placer* ids will be used by the
                building block.

                #. `placer_ids` is passed to the initializer: the
                   passed *placer* ids will be used by the building
                   block.

                #. `placer_ids` is ``None`` and the building block has
                   functional groups: The *placer* ids of the
                   functional groups will be used as the *placer* ids
                   of the building block.

                #. `placer_ids` is ``None`` and `functional_groups` is
                   empty. All atoms of the molecule will be used for
                   *placer* ids.

            position_matrix:
                The position matrix the building block should use. If
                ``None``, :func:`rdkit.ETKDGv2` will be used to
                calculate it.

        Returns:

            BuildingBlock: The building block.

        Raises:

            :class:`RuntimeError`
                If embedding the molecule fails.

        """

        editable = rdkit.EditableMol(rdkit.Mol())
        for atom in molecule.get_atoms():
            rdkit_atom = rdkit.Atom(atom.get_atomic_number())
            rdkit_atom.SetFormalCharge(atom.get_charge())
            editable.AddAtom(rdkit_atom)

        for bond in molecule.get_bonds():
            editable.AddBond(
                beginAtomIdx=bond.get_atom1_id(),
                endAtomIdx=bond.get_atom2_id(),
                order=rdkit.BondType(bond.get_order()),
            )

        rdkit_molecule = editable.GetMol()
        rdkit.SanitizeMol(rdkit_molecule)
        rdkit_molecule = rdkit.AddHs(rdkit_molecule)

        if position_matrix is None:
            params = rdkit.ETKDGv2()
            random_seed = 4
            params.randomSeed = random_seed
            if rdkit.EmbedMolecule(rdkit_molecule, params) == -1:
                raise RuntimeError(
                    f"Embedding with seed value of {random_seed} failed."
                )
        else:
            # Make sure the position matrix always holds floats.
            position_matrix = np.array(
                position_matrix,
                dtype=np.float64,
            )
            conformer = rdkit.Conformer(rdkit_molecule.GetNumAtoms())
            for atom_id, position in enumerate(position_matrix):
                conformer.SetAtomPosition(atom_id, position)
            rdkit_molecule.AddConformer(conformer)

        rdkit.Kekulize(rdkit_molecule)
        return cls.init_from_rdkit_mol(
            molecule=rdkit_molecule,
            functional_groups=functional_groups,
            placer_ids=placer_ids,
        )

    @classmethod
    def init(
        cls,
        atoms: Iterable[Atom],
        bonds: Iterable[Bond],
        position_matrix: np.ndarray,
        functional_groups: (
            FunctionalGroup
            | FunctionalGroupFactory
            | Iterable[FunctionalGroup | FunctionalGroupFactory]
        ) = (),
        placer_ids: Iterable[int] | None = None,
    ) -> typing.Self:
        """
        Initialize a :class:`.BuildingBlock` from its components.

        Parameters:

            atoms (list[Atom]):
                The atoms of the building block.

            bonds (list[Bond]):
                The bonds of the building block.

            position_matrix:
                An ``(n, 3)`` position matrix of the building block.

            functional_groups (FunctionalGroup \
| FunctionalGroupFactory \
| list[FunctionalGroup | FunctionalGroupFactory]):
                :class:`.FunctionalGroup` instances added to the
                building block and :class:`.FunctionalGroupFactory`
                instances used to create :class:`.FunctionalGroup`
                instances added to the building block.
                :class:`.FunctionalGroup` instances are used to
                identify which atoms are modified during
                :class:`.ConstructedMolecule` construction.

            placer_ids (list[int]):
                The ids of *placer* atoms. These are the atoms which
                should be used for calculating the position of the
                building block. Depending on the values passed to
                `placer_ids`, and the functional groups in the building
                block, different *placer* ids will be used by the
                building block.

                #. `placer_ids` is passed to the initializer: the
                   passed *placer* ids will be used by the building
                   block.

                #. `placer_ids` is ``None`` and the building block has
                   functional groups: The *placer* ids of the
                   functional groups will be used as the *placer* ids
                   of the building block.

                #. `placer_ids` is ``None`` and `functional_groups` is
                   empty. All atoms of the molecule will be used for
                   *placer* ids.

        Returns:

            BuildingBlock: The building block.

        """

        building_block = cls.__new__(cls)
        Molecule.__init__(
            self=building_block,
            atoms=atoms,
            bonds=bonds,
            position_matrix=position_matrix,
        )
        functional_groups = building_block._extract_functional_groups(
            functional_groups=functional_groups,
        )
        building_block._with_functional_groups(functional_groups)
        building_block._fg_repr = repr(  # type: ignore[has-type]
            tuple(building_block.get_functional_groups())
        )
        building_block._placer_ids = building_block._normalize_placer_ids(
            placer_ids=placer_ids,
            functional_groups=building_block._functional_groups,
        )
        building_block._core_ids = frozenset(
            building_block._get_core_ids(
                functional_groups=building_block._functional_groups,
            )
        )
        return building_block

    @classmethod
    def init_from_file(
        cls,
        path: pathlib.Path | str,
        functional_groups: (
            FunctionalGroup
            | FunctionalGroupFactory
            | Iterable[FunctionalGroup | FunctionalGroupFactory]
        ) = (),
        placer_ids: Iterable[int] | None = None,
    ) -> typing.Self:
        """
        Initialize from a file.

        Parameters:

            path:
                The path to a molecular structure file. Supported file
                types are:

                    #. ``.mol``, ``.sdf`` - MDL V3000 MOL file

            functional_groups (FunctionalGroup \
| FunctionalGroupFactory \
| list[FunctionalGroup | FunctionalGroupFactory]):
                :class:`.FunctionalGroup` instances added to the
                building block and :class:`.FunctionalGroupFactory`
                instances used to create :class:`.FunctionalGroup`
                instances added to the building block.
                :class:`.FunctionalGroup` instances are used to
                identify which atoms are modified during
                :class:`.ConstructedMolecule` construction.

            placer_ids (list[int]):
                The ids of *placer* atoms. These are the atoms which
                should be used for calculating the position of the
                building block. Depending on the values passed to
                `placer_ids`, and the functional groups in the building
                block, different *placer* ids will be used by the
                building block.

                #. `placer_ids` is passed to the initializer: the
                   passed *placer* ids will be used by the building
                   block.

                #. `placer_ids` is ``None`` and the building block has
                   functional groups: The *placer* ids of the
                   functional groups will be used as the *placer* ids
                   of the building block.

                #. `placer_ids` is ``None`` and `functional_groups` is
                   empty. All atoms of the molecule will be used for
                   *placer* ids.

        Returns:

            BuildingBlock: The building block.

        Raises:

            :class:`ValueError`
                If the file type cannot be used for initialization.

        """

        _, extension = os.path.splitext(path)

        if extension not in cls._init_funcs:
            raise ValueError(f'Unable to initialize from "{extension}" files.')
        # This remake needs to be here because molecules loaded
        # with rdkit often have issues, because rdkit tries to do
        # bits of structural analysis like stereocenters. Remake
        # gets rid of all this problematic metadata.
        molecule = remake(cls._init_funcs[extension](str(path)))

        return cls.init_from_rdkit_mol(
            molecule=molecule,
            functional_groups=functional_groups,
            placer_ids=placer_ids,
        )

    @classmethod
    def init_from_rdkit_mol(
        cls,
        molecule: rdkit.Mol,
        functional_groups: (
            FunctionalGroup
            | FunctionalGroupFactory
            | Iterable[FunctionalGroup | FunctionalGroupFactory]
        ) = (),
        placer_ids: Iterable[int] | None = None,
    ) -> typing.Self:
        """
        Initialize from an :mod:`rdkit` molecule.

        Warnings:

            For :mod:`rdkit` molecules with non-integer bond orders,
            such as 1.5, the molecule should be kekulized prior to
            calling this method. Otherwise, all bond orders will be
            set to an integer value in the building block.

        Parameters:

            molecule:
                The molecule.

            functional_groups (FunctionalGroup \
| FunctionalGroupFactory \
| list[FunctionalGroup | FunctionalGroupFactory]):
                :class:`.FunctionalGroup` instances added to the
                building block and :class:`.FunctionalGroupFactory`
                instances used to create :class:`.FunctionalGroup`
                instances added to the building block.
                :class:`.FunctionalGroup` instances are used to
                identify which atoms are modified during
                :class:`.ConstructedMolecule` construction.

            placer_ids (list[int]):
                The ids of *placer* atoms. These are the atoms which
                should be used for calculating the position of the
                building block. Depending on the values passed to
                `placer_ids`, and the functional groups in the building
                block, different *placer* ids will be used by the
                building block.

                #. `placer_ids` is passed to the initializer: the
                   passed *placer* ids will be used by the building
                   block.

                #. `placer_ids` is ``None`` and the building block has
                   functional groups: The *placer* ids of the
                   functional groups will be used as the *placer* ids
                   of the building block.

                #. `placer_ids` is ``None`` and `functional_groups` is
                   empty. All atoms of the molecule will be used for
                   *placer* ids.

        Returns:

            BuildingBlock: The building block.

        """

        building_block = cls.__new__(cls)
        building_block._init_from_rdkit_mol(
            molecule=molecule,
            functional_groups=functional_groups,
            placer_ids=placer_ids,
        )
        return building_block

    def with_structure_from_atomlite(
        self,
        database: atomlite.Database,
        key: str,
    ) -> typing.Self:
        """
        Return a clone, with its structure taken from an atomlite database.

        Parameters:

            database:
                The :mod:`atomlite` database.

            key:
                The key of the molecule in the database.

        Returns:

            Molecule: A clone with atomic positions found in `path`.

        Raises:

            ValueError: If the entry is not found in the database.

        """

        entry = database.get_entry(key)
        if entry is None:
            msg = f"Entry {key} not found in database."
            raise ValueError(msg)

        position_matrix = np.array(entry.molecule["conformers"][0])
        return self.clone().with_position_matrix(position_matrix)

    @classmethod
    def init_from_atomlite(
        cls,
        database: atomlite.Database,
        key: str,
        functional_groups: (
            FunctionalGroup
            | FunctionalGroupFactory
            | Iterable[FunctionalGroup | FunctionalGroupFactory]
        ) = (),
        placer_ids: Iterable[int] | None = None,
    ) -> typing.Self:
        """
        Initialize from an :mod:`atomlite` database entry.

        Parameters:

            database:
                The :mod:`atomlite` database.

            key:
                The key of the molecule in the database.

            functional_groups (FunctionalGroup \
| FunctionalGroupFactory \
| list[FunctionalGroup | FunctionalGroupFactory]):
                :class:`.FunctionalGroup` instances added to the
                building block and :class:`.FunctionalGroupFactory`
                instances used to create :class:`.FunctionalGroup`
                instances added to the building block.
                :class:`.FunctionalGroup` instances are used to
                identify which atoms are modified during
                :class:`.ConstructedMolecule` construction.

            placer_ids (list[int]):
                The ids of *placer* atoms. These are the atoms which
                should be used for calculating the position of the
                building block. Depending on the values passed to
                `placer_ids`, and the functional groups in the building
                block, different *placer* ids will be used by the
                building block.

                #. `placer_ids` is passed to the initializer: the
                   passed *placer* ids will be used by the building
                   block.

                #. `placer_ids` is ``None`` and the building block has
                   functional groups: The *placer* ids of the
                   functional groups will be used as the *placer* ids
                   of the building block.

                #. `placer_ids` is ``None`` and `functional_groups` is
                   empty. All atoms of the molecule will be used for
                   *placer* ids.

        Returns:

            BuildingBlock: The building block.

        Raises:

            ValueError: If the entry is not found in the database.

        """

        entry = database.get_entry(key)
        if entry is None:
            msg = f"Entry {key} not found in database."
            raise ValueError(msg)

        rdkit_molecule = atomlite.json_to_rdkit(entry.molecule)
        building_block = cls.__new__(cls)
        building_block._init_from_rdkit_mol(
            molecule=rdkit_molecule,
            functional_groups=functional_groups,
            placer_ids=placer_ids,
        )
        return building_block

    def _init_from_rdkit_mol(
        self,
        molecule: rdkit.Mol,
        functional_groups: (
            FunctionalGroup
            | FunctionalGroupFactory
            | Iterable[FunctionalGroup | FunctionalGroupFactory]
        ),
        placer_ids: Iterable[int] | None,
    ) -> None:
        """
        Initialize from an :mod:`rdkit` molecule.

        Parameters:

            molecule:
                The molecule.

            functional_groups (FunctionalGroup \
| FunctionalGroupFactory \
| list[FunctionalGroup | FunctionalGroupFactory]):
                :class:`.FunctionalGroup` instances added to the
                building block and :class:`.FunctionalGroupFactory`
                instances used to create :class:`.FunctionalGroup`
                instances added to the building block.
                :class:`.FunctionalGroup` instances are used to
                identify which atoms are modified during
                :class:`.ConstructedMolecule` construction.

            placer_ids (list[int]):
                The ids of *placer* atoms. These are the atoms which
                should be used for calculating the position of the
                building block. Depending on the values passed to
                `placer_ids`, and the functional groups in the building
                block, different *placer* ids will be used by the
                building block.

                #. `placer_ids` is passed to the initializer: the
                   passed *placer* ids will be used by the building
                   block.

                #. `placer_ids` is ``None`` and the building block has
                   functional groups: The *placer* ids of the
                   functional groups will be used as the *placer* ids
                   of the building block.

                #. `placer_ids` is ``None`` and `functional_groups` is
                   empty. All atoms of the molecule will be used for
                   *placer* ids.

        """

        atoms = tuple(
            Atom(a.GetIdx(), a.GetAtomicNum(), a.GetFormalCharge())
            for a in molecule.GetAtoms()
        )
        bonds = tuple(
            Bond(
                atom1=atoms[b.GetBeginAtomIdx()],
                atom2=atoms[b.GetEndAtomIdx()],
                order=(
                    9
                    if b.GetBondType() == rdkit.BondType.DATIVE
                    else b.GetBondTypeAsDouble()
                ),
            )
            for b in molecule.GetBonds()
        )
        position_matrix = molecule.GetConformer().GetPositions()

        super().__init__(atoms, bonds, position_matrix)
        self._with_functional_groups(
            self._extract_functional_groups(
                functional_groups=functional_groups,
            )
        )
        self._fg_repr = repr(self._functional_groups)
        self._placer_ids = self._normalize_placer_ids(
            placer_ids=placer_ids,
            functional_groups=self._functional_groups,
        )
        self._core_ids = frozenset(
            self._get_core_ids(
                functional_groups=self._functional_groups,
            )
        )

    def _normalize_placer_ids(
        self,
        placer_ids: Iterable[int] | None,
        functional_groups: Collection[FunctionalGroup],
    ) -> frozenset[int]:
        """
        Get the final *placer* ids.

        Parameters:

            placer_ids: The ids of *placer* atoms or ``None``.

            functional_groups:
                The :class:`.FunctionalGroup` instances of the building
                block.

        Returns:

            Depending on the input values, this function will return
            different things.

            #. `placer_ids` is a :class:`tuple` of :class`int`: the
                `placer_ids` will be returned.

            #. `placer_ids` is ``None`` and `functional_groups` is not
                empty: The *placer* ids of the functional groups will
                be returned.

            #. `placer_ids` is ``None`` and `functional_groups` is
               empty. The ids of all atoms in the building block will
               be returned.

        """

        if placer_ids is not None:
            return frozenset(placer_ids)

        if functional_groups:
            return frozenset(
                flatten(
                    functional_group.get_placer_ids()
                    for functional_group in functional_groups
                )
            )

        return frozenset(atom.get_id() for atom in self._atoms)

    def _get_core_ids(
        self,
        functional_groups: Iterable[FunctionalGroup],
    ) -> Iterator[int]:
        """
        Get the final *core* ids.

        This method may return duplicate ids.

        Parameters:

            functional_groups:
                The :class:`.FunctionalGroup` instances of the building
                block.

        Yields:

            The id of an atom defining the core of the molecule.

        """

        functional_group_atom_ids = {
            atom_id
            for functional_group in functional_groups
            for atom_id in functional_group.get_atom_ids()
        }
        for atom in self._atoms:
            atom_id = atom.get_id()
            if atom_id not in functional_group_atom_ids:
                yield atom_id

        for functional_group in functional_groups:
            for atom_id in functional_group.get_core_atom_ids():
                yield atom_id

    def _extract_functional_groups(
        self,
        functional_groups: (
            FunctionalGroup
            | FunctionalGroupFactory
            | Iterable[FunctionalGroup | FunctionalGroupFactory]
        ),
    ) -> Iterator[FunctionalGroup]:
        """
        Yield functional groups.

        The input can be a mixture of :class:`.FunctionalGroup` and
        :class:`.FunctionalGroupFactory`. The output yields
        :class:`.FunctionalGroup` instances only. Either those
        held directly in `functional_groups` or created by the
        factories in `functional_groups`.

        Parameters:

            functional_groups:
                Can be an :class:`iterable` of both
                :class:`.FunctionalGroup` and
                :class:`.FunctionalGroupFactory`.

        Yields:

            A functional group from `functional_groups`, or created
            by a factory in `functional_groups`.

        """

        if isinstance(
            functional_groups,
            FunctionalGroup | FunctionalGroupFactory,
        ):
            functional_groups = (functional_groups,)

        for functional_group in functional_groups:
            if isinstance(functional_group, FunctionalGroup):
                yield functional_group
            else:
                # Else it's a factory.
                yield from functional_group.get_functional_groups(self)

    def _with_functional_groups(
        self,
        functional_groups: Iterable[FunctionalGroup],
    ) -> typing.Self:
        """
        Modify the molecule.

        """

        self._functional_groups = tuple(functional_groups)
        self._fg_repr = repr(self._functional_groups)
        return self

    def with_functional_groups(
        self,
        functional_groups: Iterable[FunctionalGroup],
    ) -> typing.Self:
        """
        Return a clone with specific functional groups.

        Parameters:

            functional_groups (list[FunctionalGroup]):
                Functional groups the clone should have.

        Returns:

            BuildingBlock: The clone.

        """

        return self.clone()._with_functional_groups(functional_groups)

    def _with_canonical_atom_ordering(self) -> typing.Self:
        ordering = rdkit.CanonicalRankAtoms(self.to_rdkit_mol())
        super()._with_canonical_atom_ordering()
        id_map = {old_id: new_id for old_id, new_id in enumerate(ordering)}
        self._functional_groups = tuple(
            functional_group.with_ids(id_map)
            for functional_group in self._functional_groups
        )
        self._fg_repr = repr(self._functional_groups)
        self._placer_ids = frozenset(
            id_map[placer_id] for placer_id in self._placer_ids
        )
        self._core_ids = frozenset(
            id_map[core_id] for core_id in self._core_ids
        )
        return self

    def get_num_functional_groups(self) -> int:
        """
        Return the number of functional groups.

        Returns:

            The number of functional groups in the building block.

        """

        return len(self._functional_groups)

    def get_functional_groups(
        self,
        fg_ids: int | Iterable[int] | None = None,
    ) -> Iterator[FunctionalGroup]:
        """
        Yield the functional groups, ordered by id.

        Parameters:

            fg_ids (int | list[int] | None):
                The ids of functional groups yielded. If ``None``, then
                all functional groups are yielded.

        Yields:

            A functional group of the building block.

        """

        if fg_ids is None:
            fg_ids = range(len(self._functional_groups))
        elif isinstance(fg_ids, int):
            fg_ids = (fg_ids,)

        for fg_id in fg_ids:
            yield self._functional_groups[fg_id]

    def clone(self) -> typing.Self:
        """
        Return a clone.

        Returns:
            BuildingBlock: The clone.
        """
        clone = super().clone()
        clone._fg_repr = self._fg_repr
        clone._functional_groups = self._functional_groups
        clone._placer_ids = self._placer_ids
        clone._core_ids = self._core_ids
        return clone

    def get_num_placers(self) -> int:
        """
        Return the number of *placer* atoms in the building block.

        Returns:

            The number of *placer* atoms in the building block.

        """

        return len(self._placer_ids)

    def get_placer_ids(self) -> Iterator[int]:
        """
        Yield the ids of *placer* atoms.

        *Placer* atoms are those, which should be used to calculate
        the position of the building block.

        See Also:

            * :meth:`.FunctionalGroup.get_placer_ids`: For getting
              the placer ids of functional groups.

        Yields:

            The id of a *placer* atom.

        """

        yield from self._placer_ids

    def get_core_atom_ids(self) -> Iterator[int]:
        """
        Yield ids of atoms which form the core of the building block.

        This includes all atoms in the building block not part of a
        functional group, as well as any atoms in a functional group,
        specifically labelled as core atoms.

        See Also:

            * :meth:`.FunctionalGroup.get_core_atom_ids`: For getting
              the core atom ids of functional groups.

        Yields:

            The id of a *core* atom.


        """

        yield from self._core_ids

    def with_canonical_atom_ordering(self) -> typing.Self:
        """
        Return a clone, with canonically ordered atoms.

        Returns:
            BuildingBlock: The clone.
        """
        return super().with_canonical_atom_ordering()

    def with_centroid(
        self,
        position: np.ndarray,
        atom_ids: int | Iterable[int] | None = None,
    ) -> typing.Self:
        """
        Return a clone with its centroid at `position`.

        Parameters:

            position:
                This array holds the position on which the centroid of
                the clone is going to be placed.

            atom_ids (int | list[int] | None):
                The ids of atoms which should have their centroid set
                to `position`. If ``None``, all atoms are used.

        Returns:

            BuildingBlock: A clone with its centroid at `position`.
        """
        return super().with_centroid(position, atom_ids)

    def with_displacement(self, displacement: np.ndarray) -> typing.Self:
        """
        Return a displaced clone.

        Parameters:

            displacement:
                The displacement vector to be applied.

        Returns:

            BuildingBlock: A displaced clone.
        """
        return super().with_displacement(displacement)

    def with_position_matrix(self, position_matrix: np.ndarray) -> typing.Self:
        """
        Return a clone with atomic positions set by `position_matrix`.

        Parameters:

            position_matrix:
                The position matrix of the clone. The shape of the
                matrix is ``(n, 3)``.

        Returns:

            BuildingBlock: The clone.

        """
        return super().with_position_matrix(position_matrix)

    def with_rotation_about_axis(
        self,
        angle: float,
        axis: np.ndarray,
        origin: np.ndarray,
    ) -> typing.Self:
        """
        Return a rotated clone.

        The clone is rotated by `angle` about `axis` on the
        `origin`.

        Parameters:

            angle:
                The size of the rotation in radians.

            axis:
                The axis about which the rotation happens. Must have
                unit magnitude.

            origin:
                The origin about which the rotation happens.

        Returns:

            BuildingBlock: A rotated clone.

        """
        return super().with_rotation_about_axis(angle, axis, origin)

    def with_rotation_between_vectors(
        self,
        start: np.ndarray,
        target: np.ndarray,
        origin: np.ndarray,
    ) -> typing.Self:
        """
        Return a rotated clone.

        The rotation is equal to a rotation from `start` to `target`.

        Given two direction vectors, `start` and `target`, this method
        applies the rotation required transform `start` to `target`
        onto the clone. The rotation occurs about the `origin`.

        For example, if the `start` and `target` vectors
        are 45 degrees apart, a 45 degree rotation will be applied to
        the clone. The rotation will be along the appropriate
        direction.

        The great thing about this method is that you as long as you
        can associate a geometric feature of the molecule with a
        vector, then the clone can be rotated so that this vector is
        aligned with `target`. The defined vector can be virtually
        anything. This means that any geometric feature of the molecule
        can be easily aligned with any arbitrary direction.

        Parameters:

            start:
                A vector which is to be rotated so that it transforms
                into the `target` vector.

            target:
                The vector onto which `start` is rotated.

            origin:
                The point about which the rotation occurs.

        Returns:

            BuildingBlock: A rotated clone.

        """
        return super().with_rotation_between_vectors(start, target, origin)

    def with_rotation_to_minimize_angle(
        self,
        start: np.ndarray,
        target: np.ndarray,
        axis: np.ndarray,
        origin: np.ndarray,
    ) -> typing.Self:
        """
        Return a rotated clone.

        The clone is rotated by the rotation required to minimize
        the angle between `start` and `target`.

        Note that this function will not necessarily overlay the
        `start` and `target` vectors. This is because the possible
        rotation is restricted to the `axis`.

        Parameters:

            start:
                The vector which is rotated.

            target:
                The vector which is stationary.

            axis:
                The vector about which the rotation happens. Must have
                unit magnitude.

            origin:
                The origin about which the rotation happens.

        Returns:

            BuildingBlock: A rotated clone.

        Raises:

            :class:`ValueError`
                If `target` has a magnitude of 0. In this case it is
                not possible to calculate an angle between `start` and
                `target`.

        """
        return super().with_rotation_to_minimize_angle(
            start=start,
            target=target,
            axis=axis,
            origin=origin,
        )

    def with_structure_from_file(
        self,
        path: pathlib.Path | str,
        extension: str | None = None,
    ) -> typing.Self:
        """
        Return a clone, with its structure taken from a file.

        Multiple file types are supported, namely:

        #. ``.mol``, ``.sdf`` - MDL V2000 and V3000 files
        #. ``.xyz`` - XYZ files
        #. ``.mae`` - Schrodinger Maestro files
        #. ``.coord`` - Turbomole files
        #. ``.pdb`` - PDB files

        Parameters:

            path:
                The path to a molecular structure file holding updated
                coordinates for the :class:`.Molecule`.

            extension:
                If you want to treat the file as though it has a
                particular extension, put it here. Include the dot.

        Returns:

            BuildingBlock: A clone with atomic positions found in `path`.

        """
        return super().with_structure_from_file(str(path), extension)

    def write(
        self,
        path: pathlib.Path | str,
        atom_ids: int | Iterable[int] | None = None,
    ) -> typing.Self:
        """
        Write the structure to a file.

        This function will write the format based on the extension
        of `path`.

        #. ``.mol``, ``.sdf`` - MDL V3000 MOL file
        #. ``.xyz`` - XYZ file
        #. ``.pdb`` - PDB file

        Parameters:

            path:
                The `path` to which the molecule should be written.

            atom_ids (int | list[int] | None):
                The ids of atoms to write. If ``None``, all
                atoms are used. If you use this parameter, the atom
                ids in the file may not correspond to the atom ids
                in the molecule.

        Returns:

            BuildingBlock: The molecule.

        """
        return super().write(path, atom_ids)

    def __str__(self) -> str:
        smiles = rdkit.MolToSmiles(
            mol=rdkit.RemoveHs(self.to_rdkit_mol()),
        )
        return f"{self.__class__.__name__}({smiles!r}, {self._fg_repr})"

    def __repr__(self) -> str:
        return str(self)
